# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import BLOCK_SIZE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.tape import Tape
from ..uic import ANY_UIC, DEFAULT_UIC, UIC
from .dos11fs import (
    DEFAULT_PROTECTION_CODE,
    date_to_dos11,
    dos11_canonical_filename,
    dos11_split_fullname,
    dos11_to_date,
)
from .rad50 import asc_to_rad50_word, rad50_word_to_asc

__all__ = [
    "DOS11MagTapeFile",
    "DOS11MagTapeDirectoryEntry",
    "DOS11MagTapeFilesystem",
]

HEADER_RECORD = "<HHHHHHH"
HEADER_RECORD_SIZE = 14
RECORD_SIZE = 512


class DOS11MagTapeFile(AbstractFile):
    entry: "DOS11MagTapeDirectoryEntry"
    closed: bool
    size: int  # size in bytes
    content: bytes  # file content

    def __init__(self, entry: "DOS11MagTapeDirectoryEntry"):
        self.entry = entry
        self.closed = False
        self.size = entry.size
        entry.fs.dev.tape_seek(entry.tape_pos)
        self.content = entry.fs.dev.tape_read_file()[HEADER_RECORD_SIZE:]

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        return self.content[block_number * BLOCK_SIZE : (block_number + number_of_blocks) * BLOCK_SIZE]

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class DOS11MagTapeDirectoryEntry(AbstractDirectoryEntry):
    """
    DOS-11 MagTape File Header

        +-------------------------------------+
     0  |               File                  |
     2  |               name                  |
        +-------------------------------------+
     4  |            Extension                |
        +-------------------------------------+
     6  |               UIC                   |
        +-------------------------------------+
     8  |          Protection code            |
        +-------------------------------------+
    10  |           Creation Date             |
        +-------------------------------------+
    12  |             File name               |
        +-------------------------------------+
    """

    fs: "DOS11MagTapeFilesystem"
    uic: UIC = DEFAULT_UIC
    filename: str = ""
    extension: str = ""
    raw_creation_date: int = 0
    protection_code: int = 0  # System Programmers Manual, Pag 140
    size: int = 0  # size in bytes
    tape_pos: int = 0  # tape position (before file header)

    def __init__(self, fs: "DOS11MagTapeFilesystem"):
        self.fs = fs

    @classmethod
    def new(
        cls,
        fs: "DOS11MagTapeFilesystem",
        tape_pos: int,
        uic: UIC,
        filename: str,
        extension: str,
        creation_date: t.Optional[date] = None,  # optional creation date
        protection_code: int = DEFAULT_PROTECTION_CODE,
        size: int = 0,
    ) -> "DOS11MagTapeDirectoryEntry":
        self = DOS11MagTapeDirectoryEntry(fs)
        self.tape_pos = tape_pos
        self.uic = uic
        self.filename = filename
        self.extension = extension
        self.raw_creation_date = date_to_dos11(creation_date) if creation_date is not None else 0
        self.protection_code = protection_code
        self.size = size
        return self

    @classmethod
    def read(
        cls,
        fs: "DOS11MagTapeFilesystem",
        buffer: bytes,
        tape_pos: int,
        size: int,
    ) -> "DOS11MagTapeDirectoryEntry":
        self = DOS11MagTapeDirectoryEntry(fs)
        self.tape_pos = tape_pos
        (
            fnam1,
            fnam2,
            ftyp,
            fuic,
            self.protection_code,
            self.raw_creation_date,
            fnam3,
        ) = struct.unpack_from(HEADER_RECORD, buffer, 0)
        self.filename = rad50_word_to_asc(fnam1) + rad50_word_to_asc(fnam2) + rad50_word_to_asc(fnam3)  # RAD50 chars
        self.extension = rad50_word_to_asc(ftyp)  # RAD50 chars
        self.uic = UIC.from_word(fuic)
        self.size = size - HEADER_RECORD_SIZE
        return self

    def write(self, skip_file: bool = True) -> bool:
        """
        Write the directory entry
        """
        buffer = bytearray(HEADER_RECORD_SIZE)
        # Convert filename to RAD50 words
        fnam1 = asc_to_rad50_word(self.filename[:3])
        fnam2 = asc_to_rad50_word(self.filename[3:6])
        fnam3 = asc_to_rad50_word(self.filename[6:9])
        ftyp = asc_to_rad50_word(self.extension)
        fuic = self.uic.to_word()
        # Pack the data into the buffer
        struct.pack_into(
            HEADER_RECORD, buffer, 0, fnam1, fnam2, ftyp, fuic, self.protection_code, self.raw_creation_date, fnam3
        )
        self.fs.dev.tape_seek(self.tape_pos)
        self.fs.dev.tape_write_forward(buffer)
        if skip_file:
            self.fs.dev.tape_skip_file()
        return True

    @property
    def length(self) -> int:
        """Length in blocks"""
        return int(math.ceil(self.size / BLOCK_SIZE))

    @property
    def is_empty(self) -> bool:
        return self.filename == "" and self.extension == ""

    @property
    def fullname(self) -> str:
        return f"{self.uic or ''}{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return int(math.ceil(self.size / BLOCK_SIZE))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    @property
    def creation_date(self) -> t.Optional[date]:
        return dos11_to_date(self.raw_creation_date)

    def delete(self) -> bool:
        """
        Delete the file
        """
        self.uic = DEFAULT_UIC
        self.filename = ""
        self.extension = ""
        self.raw_creation_date = 0
        self.protection_code = 0
        self.write()
        return True

    def open(self, file_mode: t.Optional[str] = None) -> DOS11MagTapeFile:
        """
        Open a file
        """
        return DOS11MagTapeFile(self)

    def __str__(self) -> str:
        return (
            f"{self.filename:>9}."
            f"{self.extension:<3} "
            f"{self.uic.to_wide_str() if self.uic else '':<9}  "
            f"<{self.protection_code:o}> "
            f"{self.creation_date or '          '} "
            f"{self.size:>6} "
        )

    def __repr__(self) -> str:
        return str(self)


class DOS11MagTapeFilesystem(AbstractFilesystem):
    """
    DOS/BATCH MagTape Filesystem

    All files on magnetic tape have the following format:

    Record
        +-------------------------------------+
     1  |           File header               |  14 bytes
        +-------------------------------------+
     2  |              Data                   | 512 bytes
        +-------------------------------------+
        |                                     |
     n  /               ...                   /
        |                                     |
        +-------------------------------------+
    n-1 |              Data                   | <= 512 bytes (last data block)
        +-------------------------------------+
     n  |               EOF                   |
        +-------------------------------------+

    DOS/BATCH File Utility Package (PIP), Pag 50
    http://bitsavers.informatik.uni-stuttgart.de/pdf/dec/pdp11/dos-batch/V9/DEC-11-UPPA-A-D_PIP_Aug73.pdf
    """

    fs_name = "magtape"
    fs_description = "PDP-11 DOS/BATCH Magtape"
    fs_platforms = ["pdp11"]
    dev: Tape

    uic: UIC = DEFAULT_UIC  # current User Identification Code

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = Tape(file_or_device)
        elif isinstance(file_or_device, Tape):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "DOS11MagTapeFilesystem":
        """
        Mount the filesystem from a file or device
        """
        self = cls(file_or_dev)
        # if strict:
        #     self.dev.tape_rewind()
        #     tape_pos = self.dev.tape_pos
        #     header, size = self.tape_read_header()
        #     entry = DOS11MagTapeDirectoryEntry.read(self, header, tape_pos, size)
        #     # if entry is not None or entry.size < 0:
        #     #     raise OSError(errno.EIO, os.strerror(errno.EIO))
        return self

    def read_file_headers(self, uic: UIC = ANY_UIC) -> t.Iterator["DOS11MagTapeDirectoryEntry"]:
        """Read file headers"""
        self.dev.tape_rewind()
        try:
            while True:
                tape_pos = self.dev.tape_pos
                header, size = self.dev.tape_read_header()
                if header:
                    entry = DOS11MagTapeDirectoryEntry.read(self, header, tape_pos, size)
                    if uic.match(entry.uic):
                        yield entry
        except EOFError:
            pass

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
        uic: t.Optional[UIC] = None,
    ) -> t.Iterator["DOS11MagTapeDirectoryEntry"]:
        if uic is None:
            uic = self.uic
        uic, pattern = dos11_split_fullname(fullname=pattern, wildcard=wildcard, uic=uic)
        for entry in self.read_file_headers(uic=uic):
            if filename_match(entry.basename, pattern, wildcard):
                if include_all or not entry.is_empty:
                    yield entry

    @property
    def entries_list(self) -> t.Iterator["DOS11MagTapeDirectoryEntry"]:
        for entry in self.read_file_headers(uic=self.uic):
            if not entry.is_empty:
                yield entry

    def get_file_entry(self, fullname: str) -> DOS11MagTapeDirectoryEntry:
        fullname = dos11_canonical_filename(fullname)
        if not fullname:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        uic, basename = dos11_split_fullname(fullname=fullname, wildcard=False, uic=self.uic)
        try:
            return next(self.filter_entries_list(basename, uic=uic, wildcard=False))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
        protection_code: int = DEFAULT_PROTECTION_CODE,
    ) -> None:
        """
        Write content to a file
        """
        number_of_blocks = int(math.ceil(len(content) * 1.0 / RECORD_SIZE))
        self.create_file(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            content=content,
            protection_code=protection_code,
        )

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        content: t.Optional[bytes] = None,
        protection_code: int = DEFAULT_PROTECTION_CODE,
    ) -> t.Optional[DOS11MagTapeDirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        # Delete the existing file
        uic, basename = dos11_split_fullname(fullname=fullname, wildcard=False, uic=self.uic)
        try:
            self.get_file_entry(basename).delete()  # type: ignore
        except FileNotFoundError:
            pass
        # Find the position for the new file
        tape_pos = self.dev.tape_pos - 4  # tape mark size
        self.dev.tape_truncate()
        # Create the new directory entry
        filename, extension = basename.split(".", 1)  # type: ignore
        entry = DOS11MagTapeDirectoryEntry.new(
            fs=self,
            tape_pos=tape_pos,
            uic=uic,
            filename=filename,
            extension=extension,
            creation_date=creation_date,
            protection_code=protection_code,
        )
        entry.write(skip_file=False)
        # Write the file
        empty_record = b"\0" * RECORD_SIZE
        for i in range(0, number_of_blocks):
            if content is not None:
                record = content[i * RECORD_SIZE : (i + 1) * RECORD_SIZE]
                if len(record) < RECORD_SIZE:
                    record += b"\0" * (RECORD_SIZE - len(record))
                self.dev.tape_write_forward(record)
            else:
                self.dev.tape_write_forward(empty_record)
        # Write tape mark
        self.dev.tape_write_mark()
        self.dev.tape_write_mark()
        self.dev.tape_truncate()
        return entry

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        if options.get("uic"):
            # Listing of all UIC
            sys.stdout.write(f"{volume_id}:\n\n")
            for uic in sorted(set([x.uic for x in self.read_file_headers(uic=ANY_UIC)])):
                sys.stdout.write(f"{uic.to_wide_str()}\n")
            return
        i = 0
        files = 0
        blocks = 0
        uic, pattern = dos11_split_fullname(fullname=pattern, wildcard=True, uic=self.uic)
        if not options.get("brief"):
            dt = date.today().strftime('%y-%b-%d').upper()
            sys.stdout.write(f"DIRECTORY {volume_id}: {uic}\n\n{dt}\n\n")
        for x in self.filter_entries_list(pattern, uic=uic, include_all=True):
            if x.is_empty:
                continue
            i = i + 1
            if x.is_empty:
                continue
            fullname = x.is_empty and x.filename or "%-6s.%-3s" % (x.filename, x.extension)
            if options.get("brief"):
                # Lists only file names and file types
                sys.stdout.write(f"{fullname}\n")
                continue
            creation_date = x.creation_date and x.creation_date.strftime("%d-%b-%y").upper() or ""
            attr = ""
            uic_str = x.uic.to_wide_str() if uic.has_wildcard else ""
            sys.stdout.write(
                f"{fullname:>10s} {x.length:>5d}{attr:1} {creation_date:>9s} <{x.protection_code:03o}> {uic_str}\n"
            )
            blocks += x.length
            files += 1
        if options.get("brief"):
            return
        sys.stdout.write("\n")
        sys.stdout.write(f"TOTL BLKS: {blocks:5}\n")
        sys.stdout.write(f"TOTL FILES: {files:4}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            self.dump(arg)
        else:
            sys.stdout.write("     Filename    UIC    Access Date         Size\n")
            sys.stdout.write("     --------    ---    ------ ----         -----\n")
            for entry in self.read_file_headers(uic=ANY_UIC):
                sys.stdout.write(f"{entry}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "DOS11MagTapeFilesystem":
        """
        Initialize the filesystem
        """
        self = cls(file_or_dev)
        self.dev.tape_rewind()
        self.dev.tape_write_mark()
        self.dev.tape_truncate()
        return self

    def chdir(self, fullname: str) -> bool:
        """
        Change the current User Identification Code
        """
        try:
            self.uic = UIC.from_str(fullname)
            return True
        except Exception:
            return False

    def get_pwd(self) -> str:
        return str(self.uic)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import typing as t

from ..abstract import AbstractBlockFilesystem, AbstractFile
from ..commons import TrackSector
from ..device.abstract import AbstractDevice
from .disk import AppleDisk

__all__ = [
    "AbstractAppleDiskFilesystem",
]


class AbstractAppleDiskFilesystem(AbstractBlockFilesystem):
    """Abstract base class for Apple filesystems"""

    dev: AppleDisk

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = AppleDisk(file_or_device)
        elif isinstance(file_or_device, AppleDisk):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    def read_sector(self, address: TrackSector) -> bytes:
        return self.dev.read_sector(address)

    def write_sector(self, buffer: bytes, address: TrackSector) -> None:
        self.dev.write_sector(buffer, address)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__all__ = [
    "ASCII",
    "BLOCK_SIZE",
    "DECTAPE",
    "DECTAPE_EXT",
    "IMAGE",
    "READ_FILE_FULL",
    "BlockDirection",
    "Direction",
    "TrackSector",
    "bytes_to_word",
    "dump_struct",
    "filename_match",
    "getch",
    "hex_dump",
    "pad_words",
    "pairwise",
    "splitdrive",
    "swap_words",
    "word_to_bytes",
]

import fnmatch
import sys
import typing as t
from dataclasses import dataclass
from enum import Enum

BLOCK_SIZE = 512
BYTES_PER_LINE = 16
READ_FILE_FULL = -1
ASCII = "ASCII"  # Copy in ASCII mode
IMAGE = "IMAGE"  # Copy in image mode

DECTAPE_EXT = ["tu56", "tap", "dta"]  # DECtape file extensions
DECTAPE = "dectape"  # DECtape device type


def bytes_to_word(val: bytes, position: int = 0) -> int:
    """
    Converts two bytes to a single integer (word)
    """
    return val[1 + position] << 8 | val[0 + position]


def word_to_bytes(val: int) -> bytes:
    """
    Converts an integer (word) to two bytes
    """
    return bytes([val % 256, val // 256])


def splitdrive(path: str) -> t.Tuple[str, str]:
    """
    Split a pathname into drive and path.
    """
    result = path.split(":", 1)
    if len(result) < 2:
        return ("DK", path)
    else:
        return (result[0].upper(), result[1])


def swap_words(val: int) -> int:
    """
    Swap high order and low order word in a 32-bit integer
    """
    return (val >> 16) + ((val & 0xFFFF) << 16)


def pad_words(words: t.List[int], size: int) -> t.List[int]:
    """
    Pad a list of words to a given size with zeros
    """
    return words + [0] * (size - len(words))


def hex_dump(data: bytes, bytes_per_line: int = BYTES_PER_LINE) -> None:
    """
    Display contents in hexadecimal
    """
    for i in range(0, len(data), bytes_per_line):
        line = data[i : i + bytes_per_line]
        hex_str = " ".join([f"{x:02x}" for x in line])
        ascii_str = "".join([chr(x) if 32 <= x <= 126 else "." for x in line])
        sys.stdout.write(f"{i:08x}   {hex_str.ljust(3 * bytes_per_line)}  {ascii_str}\n")


def dump_struct(d: t.Dict[str, t.Any], exclude: t.List[str] = [], include: t.List[str] = []) -> str:
    result: t.List[str] = []
    for k, v in d.items():
        if (type(v) in (int, str, bytes, list, bool) or k in include) and k not in exclude:
            if len(k) < 6:
                label = k.upper() + ":"
            else:
                label = k.replace("_", " ").title() + ":"
            result.append(f"{label:20s}{v}")
    return "\n".join(result)


def filename_match(basename: str, pattern: t.Optional[str], wildcard: bool) -> bool:
    if not pattern:
        return True
    if wildcard:
        return fnmatch.fnmatch(basename, pattern)
    else:
        return basename == pattern


try:
    import termios
    import tty

    def getch() -> str:
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

except Exception:
    import msvcrt

    def getch() -> str:
        return msvcrt.getch()  # type: ignore


class TrackSector(t.NamedTuple):
    """
    A track/sector pair
    """

    track: int
    sector: int

    def __repr__(self) -> str:
        return f"{self.track}/{self.sector}"


T = t.TypeVar('T')


def pairwise(iterator: t.Iterable[T]) -> t.Iterator[t.Tuple[T, T]]:
    """
    Iterate over an iterable in pairs.
    """
    it = iter(iterator)
    while True:
        try:
            yield next(it), next(it)
        except StopIteration:
            break


class Direction(Enum):
    """
    The direction of a block transfer
    """

    FORWARD = 0
    BACKWARD = 1

    def reverse(self) -> "Direction":
        """
        Reverse the direction
        """
        return Direction.BACKWARD if self == Direction.FORWARD else Direction.FORWARD


@dataclass
class BlockDirection:
    """
    A block number with a direction
    """

    block_number: int
    direction: Direction

    def __str__(self) -> str:
        prefix = "F" if self.direction == Direction.FORWARD else "D"
        return f"{prefix}{self.block_number}"

    def __repr__(self) -> str:
        return str(self)



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import io
import struct
import typing as t

from ..abstract import AbstractFile
from ..commons import swap_words
from ..device.abstract import AbstractDevice
from .commons import (
    Bitmap,
    InodeBitmap,
    UNIXFilesystem,
    UNIXInode,
    format_mode,
    iterate_words,
)

# Version 1 - 3

V1_SUPER_BLOCK = 0  # Superblock
V1_SUPER_BLOCK_SIZE = 2  # Superblock size in blocks

V1_USED = 0o100000  # i-node is allocated
V1_DIR = 0o040000  # directory
V1_MOD = 0o020000  # file has been modified (always on)
V1_LARGE = 0o010000  # large file

V1_SUID = 0o000040  # set user ID on execution
V1_XOWN = 0o000020  # executable
V1_ROWN = 0o000010  # read, owner
V1_WOWN = 0o000004  # write, owner
V1_ROTH = 0o000002  # read, non-owner
V1_WOTH = 0o000001  # write, non-owner

V1_PERMS = [
    [(V1_LARGE, "l"), (0, "s")],
    [(V1_DIR, "d"), (V1_SUID, "s"), (V1_XOWN, "x"), (0, "-")],
    [(V1_ROWN, "r"), (0, "-")],
    [(V1_WOWN, "w"), (0, "-")],
    [(V1_ROTH, "r"), (0, "-")],
    [(V1_WOTH, "w"), (0, "-")],
]

V1_INODE_FORMAT = "<HBBH 16s II H"
V1_FILENAME_LEN = 8
V1_INODE_SIZE = 32
V1_NADDR = 8
V1_DIR_FORMAT = f"H{V1_FILENAME_LEN}s"
V1_ROOT_INODE = 41
assert struct.calcsize(V1_INODE_FORMAT) == V1_INODE_SIZE


class UNIXInode1(UNIXInode):

    @classmethod
    def read(cls, fs: "UNIXFilesystem", inode_num: int, buffer: bytes, position: int = 0) -> "UNIXInode":
        self = UNIXInode1(fs)
        self.inode_num = inode_num
        (
            self.flags,  #   1 word  flags
            self.nlinks,  #  1 byte  number of links to file
            self.uid,  #     1 byte  user ID of owner
            self.size,  #    1 word  size
            addr,  #         8 words content or indirect block numbers
            self.ctime,  #   1 long  creation time
            self.mtime,  #   1 long  modification time
            _,  #            1 word  unused
        ) = struct.unpack_from(V1_INODE_FORMAT, buffer, position)
        self.addr = struct.unpack_from(f"{V1_NADDR}H", addr)  # type: ignore
        self.ctime = swap_words(self.ctime)  # creation time
        self.mtime = swap_words(self.mtime)  # modification tim
        return self

    def write_buffer(self, buffer: bytearray, position: int = 0) -> None:
        """
        Write the inode to the buffer
        """
        addr = struct.pack(f"{V1_NADDR}H", *self.addr)
        data = struct.pack(
            V1_INODE_FORMAT,
            self.flags,
            self.nlinks,
            self.uid,
            self.size,
            addr,
            swap_words(self.ctime),
            swap_words(self.mtime),
            0,  # unused
        )
        buffer[position : position + len(data)] = data

    def blocks(self, include_indexes: bool = False) -> t.Iterator[int]:
        if self.is_large:
            # Large file
            for block_number in self.addr:
                if block_number == 0:
                    break
                if include_indexes:
                    yield block_number
                indirect_block = self.fs.read_block(block_number)
                for n in iterate_words(indirect_block):
                    if n == 0:
                        break
                    yield n
        else:
            # Small file
            for block_number in self.addr:
                if block_number == 0:
                    break
                yield block_number

    @property
    def isdir(self) -> bool:
        return (self.flags & V1_DIR) == V1_DIR

    @property
    def is_regular_file(self) -> bool:
        return not self.isdir

    @property
    def is_large(self) -> bool:
        return bool(self.flags & V1_LARGE)

    @property
    def is_allocated(self) -> bool:
        return bool(self.flags & V1_USED)

    @property
    def is_special_file(self) -> bool:
        return self.inode_num < V1_ROOT_INODE

    def __str__(self) -> str:
        if not self.is_allocated:
            return f"{self.inode_num:>4}# ---"
        else:
            mode = format_mode(self.flags, version=self.fs.version, perms=self.fs.perms)
            return f"{self.inode_num:>4}# uid: {self.uid:>3} nlinks: {self.nlinks:>3} size: {self.size:>5} {mode} flags: {self.flags:o}"

    def examine(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Inode\n")
        buf.write(f"Inode number:         {self.inode_num}\n")
        buf.write(f"Flags:                {self.flags:06o}\n")
        if self.is_special_file:
            buf.write("Type:                 special file\n")
        elif self.isdir:
            buf.write("Type:                 directory\n")
        elif self.is_large:
            buf.write("Type:                 large file\n")
        else:
            buf.write("Type:                 file\n")
        buf.write(f"Owner user id:        {self.uid}\n")
        buf.write(f"Creation time:        {self.ctime}\n")
        buf.write(f"Modification time:    {self.mtime}\n")
        buf.write(f"Link count:           {self.nlinks}\n")
        buf.write(f"Size:                 {self.size}\n")
        if self.is_large:
            buf.write(f"Indirect blocks:      {self.addr}\n")
        buf.write(f"Blocks:               {list(self.blocks())}\n")
        return buf.getvalue()


class UNIX1Filesystem(UNIXFilesystem):
    """
    UNIX version 1, 2, 3 Filesystem

    Disk Layout

    Block
        +------------------------------+
    0   | Superblock                   |
    1   |                              |
        +------------------------------+
    2   | Inode 1 - 16                 |
        +------------------------------+
    3   | Inode 17 - 32                |
        +------------------------------+
        | ...                          |
        +------------------------------+

    Superblock layout

    Byte
         +------------------------------+
    0    | Size of free storage bitmap  |
         +------------------------------+
    2    | Bitmap                       |
         | ...                          |
         +------------------------------+
    b    | Size of i-node bitmap        |
         +------------------------------+
    b+2  | I-node bitmap                |
         | ...                          |
         +------------------------------+

    """

    fs_name = "unix1"
    fs_description = "UNIX version 1"
    fs_platforms = ["pdp-11"]
    version: int = 1  # UNIX version
    inode_size = V1_INODE_SIZE
    dir_format = V1_DIR_FORMAT
    root_inode = V1_ROOT_INODE
    first_inode = V1_ROOT_INODE
    perms = V1_PERMS
    unix_inode_class = UNIXInode1

    bitmap_size: int  # Size of free storage bitmap
    inode_bitmap_size: int  # Size of i-node bitmap

    @classmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "UNIX1Filesystem":
        self = cls(file_or_dev)
        self.pwd = "/"
        self.read_superblock()
        return self

    def read_superblock(self) -> None:
        """Read superblock"""
        superblock_data = self.read_block(V1_SUPER_BLOCK, V1_SUPER_BLOCK_SIZE)
        (self.bitmap_size,) = struct.unpack_from("<H", superblock_data, 0)
        (self.inode_bitmap_size,) = struct.unpack_from("<H", superblock_data, self.bitmap_size + 2)
        self.inodes = self.inode_bitmap_size * 8
        # bitmap = Bitmap()
        # bitmap.bitmaps = list(superblock_data[2 : self.bitmap_size + 2])
        # print(bitmap.total_bits)
        # for i in range(0, 10):
        #     print(bitmap.is_free(i+3000))
        # inode_bitmap = InodeBitmap()
        # pos = 2 + self.bitmap_size + 2
        # inode_bitmap.bitmaps = list(superblock_data[pos:pos+self.inode_bitmap_size])
        # for i in range(30, 50):
        #     print(inode_bitmap.is_free(1000+i))
        #



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import typing as t

from ..abstract import AbstractFile
from .abstract import AbstractDevice


class Tape(AbstractDevice):

    def __init__(self, file: "AbstractFile"):
        self.f = file

    @property
    def tape_pos(self) -> int:
        """
        Current tape position
        """
        return self.f.tell()

    def tape_seek(self, pos: int) -> None:
        """
        Change the tape position
        """
        self.f.seek(pos, 0)

    def tape_rewind(self) -> None:
        """
        Rewind the tape
        """
        self.f.seek(0, 0)

    def tape_read_forward(self) -> bytes:
        """
        Starting at the current position, read the next 4 bytes from the file.
        If those bytes are a valid record length, read the data record and position
        the tape past the trailing record length.
        """
        bc = self.f.read(4)
        if len(bc) == 0:
            raise EOFError
        if bc[2] != 0 or bc[3] != 0:
            raise OSError(
                errno.EIO,
                f"Invalid record size, size = 0x{bc[3]:02X}{bc[2]:02X}{bc[1]:02X}{bc[0]:02X}",
            )
        wc = (bc[1] << 8) | bc[0]
        wc = (wc + 1) & ~1
        # import sys
        # sys.stdout.write(f"{bc[3]:02X}{bc[2]:02X}{bc[1]:02X}{bc[0]:02X} record length: {wc}\n")
        if not wc:
            # Tape mark
            return b""
        buffer = self.f.read(wc)
        pad = wc - len(buffer)
        if pad:
            buffer += bytes([0] * pad)  # Pad with zeros
        bc = self.f.read(4)
        return buffer

    def tape_write_forward(self, data: bytes) -> None:
        """
        Starting at the current position, write the record length (4 bytes)
        Then write the date record and the trailing record length (4 bytes).
        """
        bc = bytearray(4)
        wc = len(data)
        bc[0] = wc & 0xFF
        bc[2] = 0
        bc[1] = (wc >> 8) & 0xFF
        bc[3] = 0
        self.f.write(bc)
        self.f.write(data)
        self.f.write(bc)

    def tape_write_mark(self) -> None:
        """
        Starting at the current position, write a tape mark marker.
        Position the tape beyond the new tape mark.
        """
        bc = bytearray(4)
        self.f.write(bc)

    def tape_read_file(self) -> bytes:
        """
        Starting at the current position, read the current file.
        """
        data = bytearray()
        while True:
            buffer = self.tape_read_forward()
            if not buffer:
                return bytes(data)
            data.extend(buffer)

    def tape_read_header(self) -> t.Tuple[bytes, int]:
        """
        Starting at the current position, read the next header and
        skip the file. Returns the length of skipped data.
        """
        header = self.tape_read_forward()
        if not header:
            return header, 0
        try:
            return header, self.tape_skip_file()
        except EOFError:
            return header, 0

    def tape_skip_file(self) -> int:
        """
        Starting at the current position, read the next 4 bytes from the file.
        If those bytes are a valid record length, position the tape past the
        trailing record length and continue until end of file occurs.
        Returns the length of skipped data.
        """
        l = 0
        while True:
            buffer = self.tape_read_forward()
            if not buffer:
                return l
            l = l + len(buffer)

    def tape_truncate(self, tape_pos: t.Optional[int] = None) -> None:
        """
        Truncate the tape file to the current position.
        """
        self.f.truncate(tape_pos if tape_pos is not None else self.tape_pos)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.f.get_size()

    def close(self) -> None:
        self.f.close()



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__all__ = [
    "UIC",
    "ANY_UIC",
    "DEFAULT_UIC",
]

ANY_USER = 0xFF
ANY_GROUP = 0xFF


class UIC:
    """
    User Identification Code
    The format of UIC if [ggg,uuu] there ggg and uuu are octal digits
    The value on the left of the comma is represents the group number,
    the value on the right represents the user's number within the group.
    """

    group: int
    user: int

    def __init__(self, group: int, user: int):
        self.group = group & 0xFF
        self.user = user & 0xFF

    @classmethod
    def from_str(cls, code_str: str) -> "UIC":
        code_str = code_str.split("[")[1].split("]")[0]
        group_str, user_str = code_str.split(",")
        if group_str == "*":
            group = ANY_GROUP
        else:
            group = int(group_str, 8) & 0xFF
        if user_str == "*":
            user = ANY_USER
        else:
            user = int(user_str, 8) & 0xFF
        return cls(group, user)

    @classmethod
    def from_word(cls, code_int: int) -> "UIC":
        group = code_int >> 8
        user = code_int & 0xFF
        return cls(group, user)

    @property
    def has_wildcard(self) -> bool:
        return self.group == ANY_GROUP or self.user == ANY_USER

    def to_word(self) -> int:
        return (self.group << 8) + self.user

    def to_wide_str(self) -> str:
        g = f"{self.group:o}" if self.group != ANY_GROUP else "*"
        u = f"{self.user:o}" if self.user != ANY_USER else "*"
        return f"[{g:>3},{u:<3}]"

    def match(self, other: "UIC") -> bool:
        if self == other:
            return True
        elif self.group == ANY_GROUP and self.user == ANY_USER:
            return True
        elif self.group == ANY_GROUP and self.user == other.user:
            return True
        elif self.group == other.group and self.user == ANY_USER:
            return True
        else:
            return False

    def __eq__(self, other: object) -> bool:
        if isinstance(other, UIC):
            return self.group == other.group and self.user == other.user
        elif isinstance(other, str):
            other_uic = UIC.from_str(other)
            return self.group == other_uic.group and self.user == other_uic.user
        elif isinstance(other, int):
            other_uic = UIC.from_word(other)
            return self.group == other_uic.group and self.user == other_uic.user
        else:
            raise ValueError("Invalid type for comparison")

    def __lt__(self, other: "UIC") -> bool:
        return self.to_word() < other.to_word()

    def __gt__(self, other: "UIC") -> bool:
        return self.to_word() > other.to_word()

    def __hash__(self) -> int:
        return hash(self.to_word())

    def __str__(self) -> str:
        g = f"{self.group:o}" if self.group != ANY_GROUP else "*"
        u = f"{self.user:o}" if self.user != ANY_USER else "*"
        return f"[{g},{u}]"

    def __repr__(self) -> str:
        return str(self)


ANY_UIC = UIC.from_str("[*,*]")
DEFAULT_UIC = UIC.from_str("[1,1]")



import random
from datetime import datetime, timedelta

import pytest

from xferx.nova.dgdosfs import (
    START_DATE,
    DGDOSFilesystem,
    date_to_rdos,
    filename_hash,
    rdos_to_date,
    swap_bytes,
)
from xferx.shell import Shell

DSK = "tests/dsk/rdos.dsk"


def test_rdos_to_date_basic():
    dt = rdos_to_date(1, 0)
    expected = START_DATE + timedelta(days=1)
    assert dt == expected

    dt = rdos_to_date(2, (3 << 8) + 15)  # 3 hours, 15 minutes
    expected = START_DATE + timedelta(days=2, hours=3, minutes=15)
    assert dt == expected

    dt = rdos_to_date(0, 0)
    assert dt is None


def test_date_to_rdos():
    dt = START_DATE + timedelta(days=5)
    days, time = date_to_rdos(dt)
    assert days == 5
    assert time == 0

    dt = START_DATE + timedelta(days=10, hours=4, minutes=20)
    days, time = date_to_rdos(dt)
    assert days == 10
    assert time == (4 << 8) + 20

    days, time = date_to_rdos(None)
    assert days == 0
    assert time == 0


def test_swap_bytes():
    original = b'\x01\x02\x03\x04'
    expected = b'\x02\x01\x04\x03'
    assert swap_bytes(original) == expected

    original = b'\x01\x02\x03'
    expected = b'\x02\x01\x03'
    assert swap_bytes(original) == expected

    original = b''
    expected = b''
    assert swap_bytes(original) == expected

    original = b'\x01'
    expected = b'\x01'
    assert swap_bytes(original) == expected

    original = b'\x00\x01\x00\x02\x00\x03\x00\x04'
    expected = b'\x01\x00\x02\x00\x03\x00\x04\x00'
    assert swap_bytes(original) == expected


def test_filename_hash():
    assert filename_hash("SYS", "DR", 0xFFFF) == 60075
    assert filename_hash("bootsys", "sv", 0xFFFF) == 35667
    assert filename_hash("MAP", "DR", 0xFFFF) == 57747


def test_date_round_trip():
    start_year = 1968
    end_year = 1999
    for i in range(0, 100):
        start_date = datetime(start_year, 1, 1)
        end_date = datetime(end_year, 12, 31, 23, 59, 59)
        random_seconds = random.randint(0, int((end_date - start_date).total_seconds()))
        random_seconds = random_seconds - (random_seconds % 60)
        dt = start_date + timedelta(seconds=random_seconds)
        (d1, t1) = date_to_rdos(dt)
        assert rdos_to_date(d1, t1) == dt


def test_bitmap():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /dgdos {DSK}", batch=True)
    fs = shell.volumes.get('T')
    bitmap = fs.read_bitmap()
    for b in fs.read_dir_entries():
        if not b.is_directory:
            for x in b.blocks():
                assert not bitmap.is_free(x)


def test_rdos():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /dgdos {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')
    assert isinstance(fs, DGDOSFilesystem)

    shell.onecmd("dir ou:", batch=True)
    shell.onecmd("dir/brief ou:", batch=True)
    # with pytest.raises(FileNotFoundError):
    #     shell.onecmd("dir/brief ou:/notfound", batch=True)
    shell.onecmd("type ou:com.cm", batch=True)

    d0 = fs.read_bytes("com.cm")
    shell.onecmd("copy ou:com.cm ou:com1.cm", batch=True)
    d1 = fs.read_bytes("com.cm")
    assert d0 == d1

    free = fs.read_bitmap().free()

    name = "SECONDPART.DR"
    entry = fs.get_file_entry(name)
    length = entry.get_length()
    assert entry.is_contiguous
    assert len(list(entry.blocks())) == length
    assert len(list(entry.blocks(include_indexes=True))) == length
    entry.delete()

    with pytest.raises(FileNotFoundError):
        fs.get_file_entry(name)

    assert free + length == fs.read_bitmap().free()



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import copy
import errno
import io
import math
import os
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile
from ..commons import (
    BLOCK_SIZE,
    READ_FILE_FULL,
    bytes_to_word,
    filename_match,
    word_to_bytes,
)
from ..device.abstract import AbstractDevice
from ..device.rx import (
    RX01_SECTOR_SIZE,
    RX01_SIZE,
    RX02_SECTOR_SIZE,
    RX02_SIZE,
    RX_SECTOR_TRACK,
)
from .abstract import AbstractRXBlockFilesystem
from .rad50 import asc2rad, rad2asc

__all__ = [
    "RT11File",
    "RT11DirectoryEntry",
    "RT11Filesystem",
]

HOMEBLK = 1
DEFAULT_DIR_SEGMENT = 6
DIR_ENTRY_SIZE = 14
DIRECTORY_SEGMENT_HEADER_SIZE = 10
DIRECTORY_SEGMENT_SIZE = BLOCK_SIZE * 2

E_TENT = 1  # Tentative file
E_MPTY = 2  # Empty area
E_PERM = 4  # Permanent file
E_EOS = 8  # End-of-segment marker
E_READ = 64  # Protected from write
E_PROT = 128  # Protected permanent file


def date_to_rt11(val: t.Optional[date]) -> int:
    """
    Translate Python date to RT-11 date
    """
    if val is None:
        return 0
    age = (val.year - 1972) // 32
    if age < 0:
        age = 0
    elif age > 3:
        age = 3
    year = (val.year - 1972) % 32
    return year + (val.day << 5) + (val.month << 10) + (age << 14)


def rt11_to_date(val: int) -> t.Optional[date]:
    """
    Translate RT-11 date to Python date
    """
    if val == 0:
        return None
    year = val & int("0000000000011111", 2)
    day = (val & int("0000001111100000", 2)) >> 5
    month = (val & int("0011110000000000", 2)) >> 10
    age = (val & int("1100000000000000", 2)) >> 14
    year = year + 1972 + age * 32
    if day == 0:
        day = 1
    if month == 0:
        month = 1
    try:
        return date(year, month, day)
    except:
        return None


def rt11_canonical_filename(fullname: t.Optional[str], wildcard: bool = False) -> str:
    """
    Generate the canonical RT11 name
    """
    fullname = (fullname or "").upper()
    try:
        filename, extension = fullname.split(".", 1)
    except Exception:
        filename = fullname
        extension = "*" if wildcard else ""
    filename = rad2asc(asc2rad(filename[0:3])) + rad2asc(asc2rad(filename[3:6]))
    extension = rad2asc(asc2rad(extension))
    return f"{filename}.{extension}"


class RT11File(AbstractFile):
    entry: "RT11DirectoryEntry"
    closed: bool
    size: int

    def __init__(self, entry: "RT11DirectoryEntry"):
        self.entry = entry
        self.closed = False
        self.size = entry.length * BLOCK_SIZE

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if self.closed or block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if block_number + number_of_blocks > self.entry.length:
            number_of_blocks = self.entry.length - block_number
        return self.entry.segment.fs.read_block(
            self.entry.file_position + block_number,
            number_of_blocks,
        )

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        self.entry.segment.fs.write_block(
            buffer,
            self.entry.file_position + block_number,
            number_of_blocks,
        )

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class RT11DirectoryEntry(AbstractDirectoryEntry):

    segment: "RT11Segment"
    type: int = 0
    clazz: int = 0
    filename: str = ""
    extension: str = ""
    length: int = 0
    job: int = 0
    channel: int = 0
    raw_creation_date: int = 0
    extra_bytes: bytes = b''
    file_position: int = 0

    def __init__(self, segment: "RT11Segment"):
        self.segment = segment

    @classmethod
    def read(
        cls,
        segment: "RT11Segment",
        buffer: bytes,
        position: int,
        file_position: int,
        extra_bytes: int,
    ) -> "RT11DirectoryEntry":
        self = cls(segment)
        self.type = buffer[position]
        self.clazz = buffer[position + 1]
        self.filename = rad2asc(buffer, position + 2) + rad2asc(buffer, position + 4)  # 6 RAD50 chars
        self.extension = rad2asc(buffer, position + 6)  # 3 RAD50 chars
        self.length = bytes_to_word(buffer, position + 8)  # length in blocks
        self.job = buffer[position + 10]
        self.channel = buffer[position + 11]
        self.raw_creation_date = bytes_to_word(buffer, position + 12)
        self.extra_bytes = buffer[position + 14 : position + 14 + extra_bytes]
        self.file_position = file_position
        return self

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.type)
        out.append(self.clazz)
        out.extend(asc2rad(self.filename[0:3]))
        out.extend(asc2rad(self.filename[3:6]))
        out.extend(asc2rad(self.extension))
        out.extend(word_to_bytes(self.length))
        out.append(self.job)
        out.append(self.channel)
        out.extend(word_to_bytes(self.raw_creation_date))
        out.extend(self.extra_bytes)
        return bytes(out)

    @property
    def is_empty(self) -> bool:
        return self.clazz & E_MPTY == E_MPTY

    @property
    def is_tentative(self) -> bool:
        return self.clazz & E_TENT == E_TENT

    @property
    def is_permanent(self) -> bool:
        return self.clazz & E_PERM == E_PERM

    @property
    def is_end_of_segment(self) -> bool:
        return self.clazz & E_EOS == E_EOS

    @property
    def is_protected_by_monitor(self) -> bool:
        return self.clazz & E_READ == E_READ

    @property
    def is_protected_permanent(self) -> bool:
        return self.clazz & E_PROT == E_PROT

    @property
    def fullname(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return self.fullname

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    @property
    def creation_date(self) -> t.Optional[date]:
        return rt11_to_date(self.raw_creation_date)

    def delete(self) -> bool:
        """
        Delete the file
        """
        # unset E_PROT,E_TENT,E_READ,E_PROT flasgs, set E_MPTY flag
        self.clazz = self.clazz & ~E_PERM & ~E_TENT & ~E_READ & ~E_PROT | E_MPTY
        self.segment.compact()
        self.segment.write()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.segment.write()
        return True

    def open(self, file_mode: t.Optional[str] = None) -> RT11File:
        """
        Open a file
        """
        return RT11File(self)

    def __str__(self) -> str:
        return (
            f"{self.fullname:<11} "
            f"{self.creation_date or '          '} "
            f"{self.length:>6} {self.type:5o} {self.clazz:5o} "
            f"{self.job:3d} {self.channel:3d} {self.file_position:6d}"
        )

    def __repr__(self) -> str:
        return str(self)


class RT11Segment(object):
    """
    Volume Directory Segment

    +--------------+
    |5-Word header |
    +--------------+
    |Entries       |
    |.             |
    |.             |
    +--------------+
    |End-of-segment|
    |Marker        |
    +--------------+
    """

    # Block number of this directory segment
    block_number = 0
    # Total number of segments in this directory (1-31)
    num_of_segments = 0
    # Segment number of the next logical directory segment
    next_logical_dir_segment = 0
    # Number of the highest segment currently in use
    highest_segment = 0
    # Number of extra bytes per directory entry
    extra_bytes = 0
    # Block number where the stored data identified by this segment begins
    data_block_number = 0
    # Max directory entries
    max_entries = 0
    # Directory entries
    entries_list: t.List["RT11DirectoryEntry"] = []

    def __init__(self, fs: "RT11Filesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "RT11Filesystem", block_number: int) -> "RT11Segment":
        """
        Read a Volume Directory Segment from disk
        """
        self = cls(fs)
        self.block_number = block_number
        t = self.fs.read_block(self.block_number, 2)
        self.num_of_segments = bytes_to_word(t, 0)
        self.next_logical_dir_segment = bytes_to_word(t, 2)
        self.highest_segment = bytes_to_word(t, 4)
        self.extra_bytes = bytes_to_word(t, 6)
        self.data_block_number = bytes_to_word(t, 8)
        self.entries_list = []

        file_position = self.data_block_number
        dir_entry_size = DIR_ENTRY_SIZE + self.extra_bytes
        self.max_entries = (DIRECTORY_SEGMENT_SIZE - DIRECTORY_SEGMENT_HEADER_SIZE) // dir_entry_size
        for position in range(DIRECTORY_SEGMENT_HEADER_SIZE, DIRECTORY_SEGMENT_SIZE - dir_entry_size, dir_entry_size):
            dir_entry = RT11DirectoryEntry.read(self, t, position, file_position, self.extra_bytes)
            file_position = file_position + dir_entry.length
            self.entries_list.append(dir_entry)
            if dir_entry.is_end_of_segment:
                break
        return self

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(word_to_bytes(self.num_of_segments))
        out.extend(word_to_bytes(self.next_logical_dir_segment))
        out.extend(word_to_bytes(self.highest_segment))
        out.extend(word_to_bytes(self.extra_bytes))
        out.extend(word_to_bytes(self.data_block_number))
        for entry in self.entries_list:
            out.extend(entry.to_bytes())
        return out + (b"\0" * (BLOCK_SIZE * 2 - len(out)))

    def write(self) -> None:
        self.fs.write_block(self.to_bytes(), self.block_number, 2)

    @property
    def next_block_number(self) -> int:
        """Block number of the next directory segment"""
        if self.next_logical_dir_segment == 0:
            return 0
        else:
            return (self.next_logical_dir_segment - 1) * 2 + self.fs.dir_segment

    def compact(self) -> None:
        """Compact multiple unused entries"""
        prev_empty_entry = None
        new_entries_list = []
        for entry in self.entries_list:
            if not entry.is_empty:
                prev_empty_entry = None
                new_entries_list.append(entry)
            elif prev_empty_entry is None:
                prev_empty_entry = entry
                new_entries_list.append(entry)
            else:
                prev_empty_entry.length = prev_empty_entry.length + entry.length
                if entry.is_end_of_segment:
                    prev_empty_entry.clazz = prev_empty_entry.clazz | E_EOS
        self.entries_list = new_entries_list

    def insert_entry_after(self, entry: "RT11DirectoryEntry", entry_number: int, length: int) -> None:
        if entry.length == length:
            return
        new_entry = copy.copy(entry)  # new empty space entry
        if entry.is_end_of_segment:
            new_entry.clazz = E_EOS
            entry.clazz = entry.clazz - E_EOS
        new_entry.length = entry.length - length
        new_entry.file_position = entry.file_position + length
        entry.length = length
        self.entries_list.insert(entry_number + 1, new_entry)
        entry.segment.write()

    def __str__(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Segment\n")
        buf.write(f"Block number:          {self.block_number}\n")
        buf.write(f"Next dir segment:      {self.next_block_number}\n")
        buf.write(f"Number of segments:    {self.num_of_segments}\n")
        buf.write(f"Highest segment:       {self.highest_segment}\n")
        buf.write(f"Max entries:           {self.max_entries}\n")
        buf.write(f"Data block:            {self.data_block_number}\n")
        buf.write("\nNum  File        Date       Length  Type Class Job Chn  Block")
        buf.write("\n---  ----        ----       ------  ---- ----- --- ---  -----\n")
        for i, x in enumerate(self.entries_list):
            buf.write(f"{i:02d}#  {x}\n")
        return buf.getvalue()


class RT11Filesystem(AbstractRXBlockFilesystem):
    """
    RT-11 Filesystem
    """

    fs_name = "rt11"
    fs_description = "PDP-11 RT-11"
    fs_platforms = ["pdp11"]

    # First directory segment block
    dir_segment: int = DEFAULT_DIR_SEGMENT
    # System version
    ver: str = ""
    # Volume Identification
    id: str = ""
    # Owner name
    owner: str = ""
    # System Identification
    sys_id: str = ""

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        **kwargs: t.Union[bool, str],
    ) -> "RT11Filesystem":
        self = cls(file_or_dev)
        self.read_home()
        return self

    def read_home(self) -> None:
        """Read home block"""
        t = self.read_block(HOMEBLK)
        self.dir_segment = bytes_to_word(t[468:470]) or DEFAULT_DIR_SEGMENT
        self.ver = rad2asc(t[470:472])
        self.id = t[472:484].decode("ascii", "replace").replace("�", "?")
        self.owner = t[484:496].decode("ascii", "replace").replace("�", "?")
        self.sys_id = t[496:508].decode("ascii", "replace").replace("�", "?")
        self.checksum = bytes_to_word(t[510:512])

    def write_home(self) -> None:
        """Write home block"""
        # Convert data to bytes
        dir_segment_bytes = word_to_bytes(self.dir_segment)
        ver_bytes = asc2rad(self.ver)
        id_bytes = self.id.encode("ascii")
        owner_bytes = self.owner.encode("ascii")
        sys_id_bytes = self.sys_id.encode("ascii")
        checksum_bytes = word_to_bytes(0)
        # Create a byte array for the home block
        home_block = bytearray([0] * BLOCK_SIZE)
        # Fill the byte array with the data
        home_block[468:470] = dir_segment_bytes
        home_block[470:472] = ver_bytes
        home_block[472:484] = id_bytes.ljust(12, b'\0')  # Pad with null bytes if needed
        home_block[484:496] = owner_bytes.ljust(12, b'\0')
        home_block[496:508] = sys_id_bytes.ljust(12, b'\0')
        home_block[510:512] = checksum_bytes
        # Write the block
        self.write_block(home_block, HOMEBLK)

    def read_dir_segments(self) -> t.Iterator["RT11Segment"]:
        """Read directory segments"""
        next_block_number = self.dir_segment
        while next_block_number != 0:
            segment = RT11Segment.read(self, next_block_number)
            next_block_number = segment.next_block_number
            yield segment

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["RT11DirectoryEntry"]:
        if pattern:
            pattern = rt11_canonical_filename(pattern, wildcard=wildcard)
        for segment in self.read_dir_segments():
            for entry in segment.entries_list:
                if filename_match(entry.basename, pattern, wildcard):
                    if not include_all and (entry.is_empty or entry.is_tentative or entry.is_end_of_segment):
                        continue
                    yield entry

    @property
    def entries_list(self) -> t.Iterator["RT11DirectoryEntry"]:
        for segment in self.read_dir_segments():
            for entry in segment.entries_list:
                yield entry

    def get_file_entry(self, fullname: str) -> RT11DirectoryEntry:  # fullname=filename+ext
        fullname = rt11_canonical_filename(fullname)
        for entry in self.entries_list:
            if entry.fullname == fullname and entry.is_permanent:
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def read_bytes(self, fullname: str, file_mode: t.Optional[str] = None) -> bytes:  # fullname=filename+ext
        entry = self.get_file_entry(fullname)
        return self.read_block(entry.file_position, entry.length)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        number_of_blocks = int(math.ceil(len(content) * 1.0 / BLOCK_SIZE))
        entry = self.create_file(fullname, number_of_blocks, creation_date, file_type)
        if not entry:
            return
        content = content + (b"\0" * BLOCK_SIZE)
        self.write_block(content, entry.file_position, entry.length)

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[RT11DirectoryEntry]:
        fullname = os.path.basename(fullname)
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        return self.allocate_space(fullname, number_of_blocks, creation_date)

    def split_segment(self, entry: RT11DirectoryEntry) -> bool:
        # entry is the last entry of the old_segment, new new segment will contain all the entries after that
        old_segment = entry.segment
        # find the new segment number
        segments = list(self.read_dir_segments())
        first_segment = segments[0]
        sn = [x.block_number for x in segments]
        p = 0
        segment_number = None
        for i in range(self.dir_segment, self.dir_segment + (first_segment.num_of_segments * 2), 2):
            p = p + 1
            if i not in sn:
                segment_number = i
                break
        if segment_number is None:
            return False
        # create the new segment
        segment = RT11Segment(self)
        segment.block_number = segment_number
        segment.num_of_segments = first_segment.num_of_segments
        segment.next_logical_dir_segment = old_segment.next_logical_dir_segment
        segment.highest_segment = 1
        segment.extra_bytes = segments[0].extra_bytes
        segment.data_block_number = entry.file_position + entry.length
        # set the next segment of the last segment
        old_segment.next_logical_dir_segment = (segment.block_number - self.dir_segment) // 2 + 1
        entry.clazz = entry.clazz | E_EOS  # entry is the last entry of the old segment
        first_segment.num_of_segments = len(segments)  # update the total num of segments
        first_segment.write()

        entry_position = -1
        for i, e in enumerate(old_segment.entries_list):
            if entry == e:
                entry_position = i
        if entry_position == 1:
            return False
        segment.entries_list = old_segment.entries_list[entry_position + 1 :]
        old_segment.entries_list = old_segment.entries_list[: entry_position + 1]
        old_segment.write()
        segment.data_block_number = entry.file_position + entry.length
        entry.clazz = entry.clazz | E_EOS
        segment.write()
        return True

    def allocate_space(
        self,
        fullname: str,  # fullname=filename+ext, length in blocks
        length: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
    ) -> RT11DirectoryEntry:
        """
        Allocate space for a new file
        """
        entry: t.Optional[RT11DirectoryEntry] = None
        entry_number: int = -1
        # Search for an empty entry to be split
        for segment in self.read_dir_segments():
            for i, e in enumerate(segment.entries_list):
                if e.is_empty and e.length >= length:
                    if entry is None or entry.length > e.length:
                        entry = e
                        entry_number = i
                        if entry.length == length:
                            break
        if entry is None:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), fullname)
        # If the entry length is equal to the requested length, don't create the new empty entity
        if entry.length != length:
            if len(entry.segment.entries_list) >= entry.segment.max_entries:
                if not self.split_segment(entry):
                    raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), fullname)
            entry.segment.insert_entry_after(entry, entry_number, length)
        # Fill the entry
        tmp = os.path.splitext(fullname.upper())
        entry.filename = tmp[0]
        entry.extension = tmp[1] and tmp[1][1:] or ""
        entry.raw_creation_date = date_to_rt11(creation_date)
        entry.job = 0
        entry.channel = 0
        if entry.is_end_of_segment:
            entry.clazz = E_PERM | E_EOS
        else:
            entry.clazz = E_PERM
        entry.length = length
        # Write the segment
        entry.segment.write()
        return entry

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        i = 0
        files = 0
        blocks = 0
        unused = 0
        for x in self.filter_entries_list(pattern, include_all=True):
            if (
                not x.is_empty
                and not x.is_tentative
                and not x.is_permanent
                and not x.is_protected_permanent
                and not x.is_protected_by_monitor
            ):
                continue
            i = i + 1
            if x.is_empty or x.is_tentative:
                if options.get("brief"):
                    continue
                fullname = "< UNUSED >"
                date = ""
                unused = unused + x.length
            else:
                fullname = x.is_empty and x.filename or "%-6s.%-3s" % (x.filename, x.extension)
                if options.get("brief"):
                    # Lists only file names and file types
                    sys.stdout.write(f"{fullname}\n")
                    continue
                date = x.creation_date and x.creation_date.strftime("%d-%b-%y") or ""
            if x.is_permanent:
                files = files + 1
                blocks = blocks + x.length
            if x.is_protected_permanent:
                attr = "P"
            elif x.is_protected_by_monitor:
                attr = "A"
            else:
                attr = " "
            sys.stdout.write("%10s %5d%1s %9s" % (fullname, x.length, attr, date))
            # sys.stdout.write(" %8d " % (x.file_position))
            if i % 2 == 1:
                sys.stdout.write("    ")
            else:
                sys.stdout.write("\n")
        if options.get("brief"):
            return
        if i % 2 == 1:
            sys.stdout.write("\n")
        sys.stdout.write(" %d Files, %d Blocks\n" % (files, blocks))
        sys.stdout.write(" %d Free blocks\n" % unused)

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            self.dump(arg)
        else:
            sys.stdout.write(f"Directory segment:     {self.dir_segment}\n")
            sys.stdout.write(f"System version:        {self.ver}\n")
            sys.stdout.write(f"Volume identification: {self.id}\n")
            sys.stdout.write(f"Owner name:            {self.owner}\n")
            sys.stdout.write(f"System identification: {self.sys_id}\n")
            for segment in self.read_dir_segments():
                sys.stdout.write(f"{segment}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "RT11Filesystem":
        """Write an RT–11 empty device directory"""
        self = cls(file_or_dev)
        size = self.dev.get_size()
        # Adjust the size for RX01/RX02 (skip track 0)
        if size == RX01_SIZE:
            size = size - RX_SECTOR_TRACK * RX01_SECTOR_SIZE
        elif size == RX02_SIZE:
            size = size - RX_SECTOR_TRACK * RX02_SECTOR_SIZE
        length = size // BLOCK_SIZE
        # Determinate the number of directory segments
        if length >= 18000:  # 9Mb
            # DW (RD51) 10Mb
            # DL (RL02) 10.4M
            # DM (RK06) 13.8M
            num_of_segments = 31
        elif length >= 4000:  # 2Mb
            # RK (RK05) 2.45M
            # DW (RD50) 5Mb
            # DL (RL01) 5.2M
            num_of_segments = 16
        elif length >= 800:  # 400Kb
            # DZ (RX50) 400K
            # DY (RX02) 512K
            num_of_segments = 4
        else:
            # DX (RX01) 256K
            num_of_segments = 1
        # Write the home block
        self.dir_segment = DEFAULT_DIR_SEGMENT
        self.ver = "V05"
        self.id = ""
        self.owner = ""
        self.sys_id = "DECRT11A"
        self.write_home()
        # Write the directory segment
        segment = RT11Segment(self)
        segment.block_number = self.dir_segment
        segment.num_of_segments = num_of_segments
        segment.next_logical_dir_segment = 0
        segment.highest_segment = 1
        segment.extra_bytes = 0
        segment.data_block_number = self.dir_segment + (num_of_segments * 2)
        # first entry
        dir_entry = RT11DirectoryEntry(segment)
        dir_entry.file_position = segment.data_block_number
        dir_entry.length = length - dir_entry.file_position
        dir_entry.clazz = 2
        dir_entry.filename = "EMPTY"
        dir_entry.extension = "FIL"
        segment.entries_list.append(dir_entry)
        # second entry
        dir_entry = RT11DirectoryEntry(segment)
        dir_entry.file_position = size
        dir_entry.clazz = 8
        segment.entries_list.append(dir_entry)
        segment.write()
        return self



# import pytest

from xferx.nova.dgdosmagtapefs import DGDOSMagTapeFilesystem
from xferx.shell import Shell

DSK = "tests/dsk/nova_magtape.tap"
DSK_DUMP = "tests/dsk/nova_magtape_dump.tap"


def test_dgdos_magtape_read():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /dgdosmt {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, DGDOSMagTapeFilesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("type t:1", batch=True)

    x = fs.read_bytes("5")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.entries_list)
    assert len(l) == 6


def test_dgdos_magtape_dump_read():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount dm0: /dgdosmt {DSK_DUMP}", batch=True)
    # fs = shell.volumes.get('T')
    # assert isinstance(fs, DGDOSMagTapeFilesystem)

    shell.onecmd("mount d: /dgdosdump dm0:5", batch=True)
    shell.onecmd("dir d:", batch=True)

    fs = shell.volumes.get('D')
    x = fs.read_bytes("Z20")
    assert x == b'\0' * 10240


#
#
# def test_dgdos_magtape_write():
#     shell = Shell(verbose=True)
#     shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
#     shell.onecmd(f"mount in: /dgdosmt {DSK}", batch=True)
#     shell.onecmd(f"mount ou: /dgdosmt {DSK}.mo", batch=True)
#     fs = shell.volumes.get('OU')
#     assert isinstance(fs, DGDOSMagTapeFilesystem)
#
#     d = fs.get_file_entry("5")
#     assert d is not None
#
#     # Delete a file
#     d.delete()
#     with pytest.raises(FileNotFoundError):
#         fs.get_file_entry("5")
#
#     # Create a file
#     shell.onecmd("copy in:5 ou:6", batch=True)
#     x2 = fs.read_bytes("6")
#     x2 = x2.rstrip(b"\0")
#     assert len(x2) == 2200
#     for i in range(0, 50):
#         assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2
#
#
# def test_dgdos_magtape_init():
#     shell = Shell(verbose=True)
#     shell.onecmd(f"mount in: /dgdosmt {DSK}", batch=True)
#     shell.onecmd(f"create /allocate:280 {DSK}.mo", batch=True)
#     shell.onecmd(f"init /dgdosmt {DSK}.mo", batch=True)
#     shell.onecmd(f"mount ou: /dgdosmt {DSK}.mo", batch=True)
#     shell.onecmd("dir ou:", batch=True)
#     shell.onecmd("copy in:* ou:", batch=True)
#
#     fs = shell.volumes.get('OU')
#     x = fs.read_bytes("5")
#     x = x.rstrip(b"\0")
#     assert len(x) == 2200
#     for i in range(0, 50):
#         assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x
#
#     # Test init mounted volume
#     shell.onecmd("init ou:", batch=True)
#     with pytest.raises(Exception):
#         fs.read_bytes("5")



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, IMAGE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.tape import Tape

__all__ = [
    "DGDOSMagTapeFile",
    "DGDOSMagTapeDirectoryEntry",
    "DGDOSMagTapeFilesystem",
]


# RDOS System Reference - Pag 35
# https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf

DATA_WORDS = 255  # data words per block
DATA_BLOCK_SIZE = DATA_WORDS * 2  # data block size in bytes
FILE_NUMBER_WORDS = 2  # extra words per block
FILE_NUMBER_SIZE = FILE_NUMBER_WORDS * 2  # file number size in bytes
TAPE_BLOCK_WORDS = DATA_WORDS + FILE_NUMBER_WORDS  # tape block size in words
TAPE_BLOCK_SIZE = TAPE_BLOCK_WORDS * 2  # tape block size in bytes
DUMP_NAME_BLOCK_ID = 0o377


def get_file_number(buffer: bytes) -> int:
    """
    Get the file number from the tape block
    """
    if len(buffer) != TAPE_BLOCK_SIZE:
        raise OSError(errno.EIO, f"Invalid block size {len(buffer)}")
    file_number1, file_number2 = struct.unpack(">HH", buffer[-FILE_NUMBER_SIZE:])
    if file_number1 != file_number2:
        raise OSError(errno.EIO, f"Invalid file number: {file_number1} != {file_number2}")
    return file_number1  # type: ignore


class DGDOSMagTapeFile(AbstractFile):
    entry: "DGDOSMagTapeDirectoryEntry"
    closed: bool
    size: int  # size in bytes

    def __init__(self, entry: "DGDOSMagTapeDirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        self.closed = False
        self.file_mode = file_mode or IMAGE
        self.size = entry.get_size()
        self._content: t.Optional[bytes] = None

    @property
    def content(self) -> bytes:
        if self._content is None:
            self.entry.fs.dev.tape_seek(self.entry.tape_pos)
            data = bytearray()
            while True:
                buffer = self.entry.fs.dev.tape_read_forward()
                if not buffer:
                    break
                if len(buffer) != TAPE_BLOCK_SIZE:
                    raise OSError(errno.EIO, os.strerror(errno.EIO))
                data.extend(buffer[:-FILE_NUMBER_SIZE])  # remove the last 2 words of eack block
            self._content = bytes(data)
        return self._content

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = self.content[block_number * DATA_BLOCK_SIZE : (block_number + number_of_blocks) * DATA_BLOCK_SIZE]
        # Convert to ASCII if needed
        if self.file_mode == ASCII:
            return bytes([0x0A if x == 0x0D else x for x in data])
        else:
            return data

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return DATA_BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class DGDOSMagTapeDirectoryEntry(AbstractDirectoryEntry):
    """
    MagTape Directory Entry
    """

    fs: "DGDOSMagTapeFilesystem"
    file_number: int  # File number
    length: int  # Length in blocks
    tape_pos: int = 0  # Tape position
    is_dump: bool = False  # Dump/raw format

    def __init__(self, fs: "DGDOSMagTapeFilesystem"):
        self.fs = fs

    @classmethod
    def read(
        cls,
        fs: "DGDOSMagTapeFilesystem",
        buffer: bytes,
        tape_pos: int,
        size: int,
    ) -> "DGDOSMagTapeDirectoryEntry":
        self = DGDOSMagTapeDirectoryEntry(fs)
        self.file_number = buffer[-1]
        self.tape_pos = tape_pos
        self.length = (len(buffer) + size) // TAPE_BLOCK_SIZE
        self.is_dump = (buffer[0] == DUMP_NAME_BLOCK_ID) if buffer else False
        return self

    @property
    def is_empty(self) -> bool:
        return False

    @property
    def fullname(self) -> str:
        return f"{self.file_number}"

    @property
    def basename(self) -> str:
        return f"{self.file_number}"

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * DATA_BLOCK_SIZE

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return DATA_BLOCK_SIZE

    @property
    def creation_date(self) -> t.Optional[date]:
        return None

    @property
    def file_type(self) -> t.Optional[str]:
        """File type"""
        return 'dump' if self.is_dump else 'raw'

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def open(self, file_mode: t.Optional[str] = None) -> DGDOSMagTapeFile:
        """
        Open a file
        """
        return DGDOSMagTapeFile(self, file_mode)

    def __str__(self) -> str:
        return f"{self.file_number:>3} {self.file_type:<4} {self.tape_pos:>10} {self.get_size():>12}"

    def __repr__(self) -> str:
        return str(self)


class DGDOSMagTapeFilesystem(AbstractFilesystem):
    """
    Data General DOS/RDOS MagTape Filesystem

    Tapes use a fixed block size of 257, 16-bit words.
    The first 255 words of each block contain user data,
    while the last two words contain the file number.

    Word
        +-------------------------------------+
     0  |                                     |
        /                Data                 / 510 bytes
    254 |                                     |
        +-------------------------------------+
    255 |             File number             | 2 bytes
        +-------------------------------------+
    256 |             File number             | 2 bytes
        +-------------------------------------+

    RDOS System Reference - Pag 35
    https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_RDOS_SystemReference_Oct83.pdf
    """

    fs_name = "dgdosmt"
    fs_description = "Data General DOS/RDOS Magtape"
    fs_platforms = ["nova"]
    dev: Tape

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = Tape(file_or_device)
        elif isinstance(file_or_device, Tape):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "DGDOSMagTapeFilesystem":
        self = cls(file_or_dev)
        if strict:
            # Check if the file is a valid tape
            self.dev.tape_rewind()
            try:
                while True:
                    # First file block
                    buffer = self.dev.tape_read_forward()
                    if not buffer:
                        break
                    file_number = get_file_number(buffer)
                    # Other file blocks
                    try:
                        while True:
                            buffer = self.dev.tape_read_forward()
                            if not buffer:
                                break
                            tmp = get_file_number(buffer)
                            if tmp != file_number:
                                raise OSError(errno.EIO, f"Invalid file number: {tmp} != {file_number}")
                    except EOFError:
                        pass
            except EOFError:
                pass
        return self

    def read_dir_entries(self) -> t.Iterator["DGDOSMagTapeDirectoryEntry"]:
        """
        Read the directory entries from the tape
        """
        self.dev.tape_rewind()
        try:
            while True:
                tape_pos = self.dev.tape_pos
                header, size = self.dev.tape_read_header()
                if not header:
                    break
                yield DGDOSMagTapeDirectoryEntry.read(self, header, tape_pos, size)
        except EOFError:
            pass

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["DGDOSMagTapeDirectoryEntry"]:
        for entry in self.read_dir_entries():
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    @property
    def entries_list(self) -> t.Iterator["DGDOSMagTapeDirectoryEntry"]:
        for entry in self.read_dir_entries():
            if not entry.is_empty:
                yield entry

    def get_file_entry(self, fullname: str) -> DGDOSMagTapeDirectoryEntry:
        try:
            return next(self.filter_entries_list(fullname, wildcard=False))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        number_of_blocks = int(math.ceil(len(content) * 1.0 / DATA_BLOCK_SIZE))
        self.create_file(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            content=content,
        )

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        content: t.Optional[bytes] = None,
    ) -> t.Optional[DGDOSMagTapeDirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        pattern = pattern.upper() if pattern else None
        if not options.get("brief"):
            sys.stdout.write("Num Type         Size\n")
            sys.stdout.write("--- ----         ----\n")
        for x in self.filter_entries_list(pattern):
            if options.get("brief"):
                sys.stdout.write(f"{x.fullname:>3}\n")
            else:
                sys.stdout.write(f"{x.fullname:>3} {x.file_type:<4} {x.get_size():>12}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            self.dump(arg)
        else:
            sys.stdout.write("Num Type   Tape pos         Size\n")
            sys.stdout.write("--- ----   --------         ----\n")
            for entry in self.read_dir_entries():
                sys.stdout.write(f"{entry}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "DGDOSMagTapeFilesystem":
        """
        Initialize the filesystem
        """
        self = cls(file_or_dev)
        self.dev.tape_rewind()
        # Logical end of tape (2 tape marks)
        self.dev.tape_write_mark()
        self.dev.tape_write_mark()
        self.dev.tape_truncate()
        return self

    def close(self) -> None:
        self.dev.close()

    def chdir(self, fullname: str) -> bool:
        return False

    def get_pwd(self) -> str:
        return ""



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import struct
import sys
import typing as t
from datetime import date, datetime, timedelta

from ..abstract import AbstractBlockFilesystem, AbstractDirectoryEntry, AbstractFile
from ..cache import BlockCache
from ..commons import BLOCK_SIZE, READ_FILE_FULL, dump_struct, filename_match
from ..device.abstract import AbstractDevice
from ..uic import ANY_GROUP, ANY_USER, UIC
from .abstract import AbstractRXBlockFilesystem
from .rad50 import asc2rad, asc_to_rad50_word, rad2asc, rad50_word_to_asc

__all__ = [
    "RSTSFile",
    "RSTSFilesystem",
]

# RSTS/E Monitor Internals, Michael Mayfield
# http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf
# RSTS/E V8.0 Internals Manual
# https://bitsavers.org/pdf/dec/pdp11/rsts_e/V08/AA-CL35A-TE_8.0intern_Sep84.pdf

# 3-level directory hierarchy:
# Master File Directory (MFD)
# Group File Directories (GFDs)
# User File Directories (UFDs)

BOOT_BLOCK = 0  # Boot block
LABEL_BLOCK_OFFSET = 0  # Label block offset from MFD/GFD (RDS1.1 and later)
GFD_POINTER_BLOCK_OFFSET = 1  # GFD pointer block offset from MFD (RDS1.1 and later)
UFD_POINTER_BLOCK_OFFSET = 1  # UFD pointer block offset from GFD (RDS1.1 and later)
GFD_NAME_ENTRY_BLOCK_OFFSET = 2  # GFD name entries pointer block (RDS1.1 and later)
DISK_PACK_LABEL_DCN = 1  # Disk pack label DCN
BLOCKETTE_FORMAT = '<HHHHHHHH'  # 8 words blockette
BLOCKETTE_LEN = struct.calcsize(BLOCKETTE_FORMAT)
DISK_PACK_LABEL_FORMAT = BLOCKETTE_FORMAT
MFD_ENTRY_FORMAT = '<HHHHBBHHH'
MFD_ENTRY_LEN = struct.calcsize(MFD_ENTRY_FORMAT)
UFD_ENTRY_FORMAT = '<HHHHBBHHH'
UFD_ENTRY_LEN = struct.calcsize(UFD_ENTRY_FORMAT)
assert MFD_ENTRY_LEN == BLOCKETTE_LEN == UFD_ENTRY_LEN
GFD_POINTER_BLOCK_FORMAT = '<255H'
UFD_POINTER_BLOCK_FORMAT = '<255H'
CLUSTER_MAP_POS = 0o760
US_UFD = 1 << 6  # USTAT bit 6 - 1 for MFD Name Entry
RDS1_FLAGS = 0o20000  # RDS1.1 or RDS1.2
RDS0_PLVL = 0  # RDS 0 - V7.x and before
RDS11_PLVL = 257  # RDS 1.1 - V8
RDS12_PLVL = 258  # RDS 1.2 - V9.0 and beyond
SAT_FILENAME = "[0,1]SATT.SYS'"  # Storage Allocation Table filename
BAD_BLOCK_FILENAME = "[0,1]BADB.SYS'"  # Bad Block filename


class PPN(UIC):
    """
    Programmer Project Number
    The format of PPN if [ggg,uuu] there ggg and uuu are decimal digits
    The value on the left of the comma is represents the project number,
    the value on the right represents the programmer's number within the project.
    """

    @classmethod
    def from_str(cls, code_str: str) -> "PPN":
        code_str = code_str.split("[")[1].split("]")[0]
        project_str, user_str = code_str.split(",")
        if project_str == "*":
            project = ANY_GROUP
        else:
            project = int(project_str) & 0xFF
        if user_str == "*":
            user = ANY_USER
        else:
            user = int(user_str) & 0xFF
        return cls(project, user)

    @classmethod
    def from_word(cls, code_int: int) -> "PPN":
        project = code_int >> 8
        user = code_int & 0xFF
        return cls(project, user)

    def to_wide_str(self) -> str:
        g = f"{self.group}" if self.group != ANY_GROUP else "*"
        u = f"{self.user}" if self.user != ANY_USER else "*"
        return f"[{g:>3},{u:<3}]"

    def __str__(self) -> str:
        g = f"{self.group}" if self.group != ANY_GROUP else "*"
        u = f"{self.user}" if self.user != ANY_USER else "*"
        return f"[{g},{u}]"

    def __repr__(self) -> str:
        return str(self)


DEFAULT_PPN = PPN.from_str("[1,1]")
ACCOUNT_1_1_PPN = PPN.from_str("[1,1]")


def rsts_to_date(udc: int, utc: int) -> datetime:
    """
    Translate RSTS/E date and time to Python datetime

    udc is (year-1970) * 100 + (day of the year)
    utc is the number of minutes before midnight
    """
    year = (udc // 1000) + 1970
    day_of_year = udc % 1000
    date = datetime(year, 1, 1) + timedelta(days=day_of_year - 1)
    time_of_day = timedelta(minutes=1440 - utc)
    full_datetime = date + time_of_day
    return full_datetime


def rsts_canonical_filename(fullname: t.Optional[str], wildcard: bool = False) -> str:
    """
    Generate the canonical RSTS/E name
    """
    fullname = (fullname or "").upper()
    try:
        filename, extension = fullname.split(".", 1)
    except Exception:
        filename = fullname
        extension = "*" if wildcard else ""
    filename = rad2asc(asc2rad(filename[0:3])) + rad2asc(asc2rad(filename[3:6]))
    extension = rad2asc(asc2rad(extension))
    return f"{filename}.{extension}"


def rsts_canonical_fullname(fullname: str, wildcard: bool = False) -> str:
    try:
        if "[" in fullname:
            ppn: t.Optional[PPN] = PPN.from_str(fullname)
            fullname = fullname.split("]", 1)[1]
        else:
            ppn = None
    except Exception:
        ppn = None
    if fullname:
        fullname = rsts_canonical_filename(fullname, wildcard=wildcard)
    return f"{ppn or ''}{fullname}"


def rsts_split_fullname(ppn: PPN, fullname: t.Optional[str], wildcard: bool = True) -> t.Tuple[PPN, t.Optional[str]]:
    if fullname:
        if "[" in fullname:
            try:
                ppn = PPN.from_str(fullname)
                fullname = fullname.split("]", 1)[1]
            except Exception:
                return ppn, fullname
        if fullname:
            fullname = rsts_canonical_filename(fullname, wildcard=wildcard)
    return ppn, fullname


class RTFSBlockCache(BlockCache):

    def __init__(self, fs: "RSTSFilesystem"):
        super().__init__(fs.dev)
        self.fs = fs

    def read_block(self, block_number: int = 0, dcn: t.Optional[int] = None) -> bytes:
        if dcn is not None:
            block_number = self.fs.dcn_to_lbn(dcn)
        return super().read_block(block_number)


class RSTSFile(AbstractFile):
    ufd_name_entry: "UFDNameEntry"
    closed: bool

    def __init__(self, ufd_name_entry: "UFDNameEntry"):
        self.ufd_name_entry = ufd_name_entry
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.ufd_name_entry.account_entry.usiz
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.ufd_name_entry.account_entry.usiz
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        cache = self.ufd_name_entry.fs.new_cache()
        cluster_dcns = self.ufd_name_entry.read_retrieval_entries(cache=cache)
        data = bytearray()
        for i in range(block_number, block_number + number_of_blocks):
            cluster = i // self.ufd_name_entry.account_entry.uclus
            cluster_block = i % self.ufd_name_entry.account_entry.uclus
            dcn = cluster_dcns[cluster] + cluster_block
            data.extend(cache.read_block(dcn=dcn))
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.ufd_name_entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.ufd_name_entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return str(self.ufd_name_entry)


class Link:
    """
    Pointer between directory entries.

    http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 39
    """

    flags: int  #   Flags
    entry: int  #   Entry within the directory block (0-31)
    cluster: int  # Cluster number within the cluster map (0-6)
    block: int  #   Block number within the directory cluster

    def __init__(self, fs: "RSTSFilesystem", ulnk: int):
        self.fs = fs
        self.flags = ulnk & 0b1111  # 4 bits
        ulnk >>= 4
        self.entry = ulnk & 0b11111  # 5 bits
        ulnk >>= 5
        self.cluster = ulnk & 0b111  # 3 bits
        ulnk >>= 3
        self.block = ulnk & 0b1111  # 4 bits

    @property
    def ulnk(self) -> int:
        return (
            (self.block & 0b1111) << 12
            | (self.cluster & 0b111) << 9
            | (self.entry & 0b11111) << 4
            | self.flags & 0b1111
        )

    @property
    def is_null(self) -> bool:
        return self.ulnk == 0

    def to_lbn(self, cluster_map: t.List[int]) -> int:
        """
        Translate the link to a Logical Block Number according to the provided cluster map
        """
        return cluster_map[self.cluster] + self.block

    def __str__(self) -> str:
        return f"{self.block:>02},{self.cluster:>01},{self.entry:>02}"

    def __repr__(self) -> str:
        return f"{self.block:>02},{self.cluster:>01},{self.entry:>02} ({self.flags:04b})"


class UFDLabelEntry:
    """
    The UFD Label Entry is the root of the UFD directory structure.

    http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 33
    """

    fs: "RSTSFilesystem"
    ulnk: Link  # Link to first name entry in UFD
    ppn: PPN  # Project Programmer Number
    ufd: str  # User File Directory label

    def __init__(self, fs: "RSTSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "RSTSFilesystem", buffer: bytes, position: int = 0) -> "UFDLabelEntry":
        self = UFDLabelEntry(fs)
        blockette = struct.unpack_from(BLOCKETTE_FORMAT, buffer, position)
        self.ulnk = Link(self.fs, blockette[0])  # Link to first name entry in UFD
        self.ppn = PPN.from_word(blockette[6])  # Project Programmer Number
        self.ufd = rad50_word_to_asc(blockette[7])
        return self

    def write_buffer(self, buffer: bytearray, position: int = 0) -> None:
        rad50_ufd = asc_to_rad50_word(self.ufd)
        struct.pack_into(
            BLOCKETTE_FORMAT,
            buffer,
            position,
            self.ulnk.ulnk,
            -1,  # Always -1 to mark this entry in use
            0,  # Unused
            0,  # Unused
            0,  # Unused
            0,  # Unused
            self.ppn.to_word(),
            rad50_ufd,
        )

    @property
    def is_ufd_label(self) -> bool:
        return self.ufd == "UFD"

    def __str__(self) -> str:
        return f"{self.ppn.to_wide_str()} ULNK: {self.ulnk} {self.ufd}"


class UFDAccountEntry:
    """
    The UFD Accounting Entry contains information about the size,
    creation date, and cluster size of the file.

    http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 35
    """

    fs: "RSTSFilesystem"
    ulnk: Link  # Link to attribute entry
    udla: int  # Date of last access
    usiz: int  # Size in blocks
    udc: int  # Creation date
    utc: int  # Creation time
    urst: str  # Runtime system name
    uclus: int  # File cluster size

    def __init__(self, fs: "RSTSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "RSTSFilesystem", buffer: bytes, position: int) -> "UFDAccountEntry":
        self = UFDAccountEntry(fs)
        (
            ulnk,  # 2 bytes Link to attribute entry
            self.udla,  # 2 bytes Date of last access
            self.usiz,  # 2 bytes Size in blocks
            self.udc,  #  2 bytes Creation date
            self.utc,  #  2 bytes Creation time
            urst1,  # 2 bytes Runtime system name (word 1)
            urst2,  # 2 bytes Runtime system name (word 2)
            self.uclus,  # 2 bytes File cluster size
        ) = struct.unpack_from(BLOCKETTE_FORMAT, buffer, position)
        self.ulnk = Link(self.fs, ulnk)
        self.urst = rad50_word_to_asc(urst1) + rad50_word_to_asc(urst2)
        return self

    def write_buffer(self, buffer: bytearray, position: int) -> None:
        urst1 = asc_to_rad50_word(self.urst[:3])
        urst2 = asc_to_rad50_word(self.urst[3:6])
        struct.pack_into(
            BLOCKETTE_FORMAT,
            buffer,
            position,
            self.ulnk.ulnk,
            self.udla,
            self.usiz,
            self.udc,
            self.utc,
            urst1,
            urst2,
            self.uclus,
        )

    def __str__(self) -> str:
        return f"USIZ: {self.usiz} UCLUS: {self.uclus:}"


class UFDNameEntry(AbstractDirectoryEntry):

    fs: "RSTSFilesystem"
    ppn: PPN = DEFAULT_PPN
    account_entry: UFDAccountEntry
    ulnk: Link  # Link to first name entry in UFD
    filename: str  # File name
    extension: str  # File type
    uprot: int  # Protection code
    ustat: int  # Status
    uacnt: int  # Access count
    uaa: Link  # Link to accounting entry
    uar: Link  # Link to the first retrieval entry

    def __init__(self, fs: "RSTSFilesystem", ppn: PPN, ufd_uar: int):
        self.fs = fs
        self.ppn = ppn
        self.ufd_uar = ufd_uar

    @classmethod
    def read(cls, fs: "RSTSFilesystem", ppn: PPN, ufd_uar: int, buffer: bytes, position: int) -> "UFDNameEntry":
        self = UFDNameEntry(fs, ppn, ufd_uar)
        (
            ulnk,  #       2 bytes  Link to first name entry in MFD
            filename1,  #  2 bytes  File name (1st word)
            filename2,  #  2 bytes  File name (2nd word)
            filetype,  #   2 bytes  File type
            self.ustat,  # 1 byte   Status
            self.uprot,  # 1 byte   Protection code
            self.uacnt,  # 2 bytes  Access count
            uaa,  #        2 bytes  Link to accounting entry
            uar,  #        2 bytes  Link to the first retrieval entry
        ) = struct.unpack_from(UFD_ENTRY_FORMAT, buffer, position)
        self.ulnk = Link(self.fs, ulnk)
        self.uaa = Link(self.fs, uaa)
        self.uar = Link(self.fs, uar)
        self.filename = rad50_word_to_asc(filename1) + rad50_word_to_asc(filename2)
        self.extension = rad50_word_to_asc(filetype)
        return self

    def write_buffer(self, buffer: bytearray, position: int) -> None:
        filename1 = asc_to_rad50_word(self.filename[:3])
        filename2 = asc_to_rad50_word(self.filename[3:6])
        filetype = asc_to_rad50_word(self.extension)
        struct.pack_into(
            UFD_ENTRY_FORMAT,
            buffer,
            position,
            self.ulnk.ulnk,
            filename1,
            filename2,
            filetype,
            self.ustat,
            self.uprot,
            self.uacnt,
            self.uaa.ulnk,
            self.uar.ulnk,
        )

    def read_retrieval_entries(self, cache: t.Optional[RTFSBlockCache] = None) -> t.List[int]:
        """
            The retrieval entries provide the information necessary to locate the file blocks on the disk.

            +-------------------------------------+
         0  |     Link to next Retrieval Entry    |
            +-------------------------------------+
         2  |          DCN of cluster 0           |
            +-------------------------------------+
            |                 ...                 |
            +-------------------------------------+
        16  |          DCN of cluster 6           |
            +-------------------------------------+

            http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 37
        """
        if cache is None:
            cache = self.fs.new_cache()

        # Read the UFD cluster map
        retrieval_entry_link = self.uar
        buffer = cache.read_block(retrieval_entry_link.block + self.ufd_uar)
        ufd_cluster_map = self.fs.read_ufd_cluster_map(buffer)

        # Read the retrieval entries
        cluster_dcns: t.List[int] = []
        while not retrieval_entry_link.is_null:
            buffer = cache.read_block(retrieval_entry_link.to_lbn(ufd_cluster_map))
            blockette = struct.unpack_from(BLOCKETTE_FORMAT, buffer, UFD_ENTRY_LEN * retrieval_entry_link.entry)
            retrieval_entry_link = Link(self.fs, blockette[0])
            cluster_dcns += blockette[1:]
        return cluster_dcns

    @property
    def fullname(self) -> str:
        return f"{self.ppn or ''}{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def creation_date(self) -> datetime:
        return rsts_to_date(self.account_entry.udc, self.account_entry.utc)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.account_entry.usiz

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def open(self, file_mode: t.Optional[str] = None) -> RSTSFile:
        """
        Open a file
        """
        return RSTSFile(self)

    @property
    def is_ufd_name_entry(self) -> bool:
        return (self.ustat & 64) == 0

    def __str__(self) -> str:
        date = self.creation_date.strftime("%d-%b-%y %H:%M")
        return f"{self.filename:<6}.{self.extension:<3} {self.account_entry.usiz:>5}   <{self.uprot:>3}> {date}  ULNK: {self.ulnk} UAA: {self.uaa} UAR: {self.uar} USTAT: {self.ustat}"


class MFDNameEntry:
    """
    The MFD Name Entry is used to catalog all the accounts.
    Each account on the disk has a MFD Name Entry associated with it.

    http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 20
    """

    fs: "RSTSFilesystem"
    ulnk: Link  # Link to first name entry in UFD
    ppn: PPN  # Project Programmer Number
    passwd: str  # Password
    uprot: int  # Protection code
    ustat: int  # Status
    uacnt: int  # Access count
    uaa: Link  # Link to accounting entry
    uar: int  # DCN of the first cluster of the user's UFD

    def __init__(self, fs: "RSTSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "RSTSFilesystem", buffer: bytes, position: int = 0) -> "MFDNameEntry":
        self = MFDNameEntry(fs)
        (
            ulnk,  #       2 bytes  Link to first name entry in MFD
            unam,  #       2 bytes  PPN project / PPN programmer
            passwd0,  #    2 bytes  Password (first word)
            passwd1,  #    2 bytes  Password (second word)
            self.ustat,  # 1 byte   Status
            self.uprot,  # 1 byte   Unused (protection code)
            self.uacnt,  # 2 bytes  Access count
            uaa,  #        2 bytes  Link to accounting entry
            self.uar,  #   2 bytes  DCN of the first cluster of the user's UFD
        ) = struct.unpack_from(MFD_ENTRY_FORMAT, buffer, position)
        self.ulnk = Link(self.fs, ulnk)
        self.uaa = Link(self.fs, uaa)
        self.ppn = PPN.from_word(unam)
        self.passwd = rad50_word_to_asc(passwd0) + rad50_word_to_asc(passwd1)
        return self

    def write_buffer(self, buffer: bytearray, position: int = 0) -> None:
        passwd0 = asc_to_rad50_word(self.passwd[:3])
        passwd1 = asc_to_rad50_word(self.passwd[3:6])
        struct.pack_into(
            MFD_ENTRY_FORMAT,
            buffer,
            position,
            self.ulnk.ulnk,
            self.ppn.to_word(),
            passwd0,
            passwd1,
            self.ustat,
            self.uprot,
            self.uacnt,
            self.uaa.ulnk,
            self.uar,
        )

    @property
    def is_mfd_name_entry(self) -> bool:
        return (self.ustat & US_UFD) != 0

    def __str__(self) -> str:
        return f"{self.ppn.to_wide_str()} {self.passwd:<9} ULNK: {self.ulnk} USTAT: {self.ustat} UACNT: {self.uacnt} UAA: {self.uaa} UAR: {self.uar:>5}"


class GFD:
    """
    Group File Directory (RDS1.1 or later)

    http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 23
    """

    fs: "RSTSFilesystem"
    mfd: "MFD"
    group: int  # Group number
    dcn: int  # DCN of GFD
    gfd_cluster_size: int  # MFD cluster size
    gfd_cluster_map: t.List[int]  # MFD cluster map (DCN of MFD clusters 0 - 6)
    ufd_pointer_map: t.List[int]  # Pointers to User File Directories
    name_entry_pointer_map: t.List[Link]  # Links to name entries

    def __init__(self, mfd: "MFD", group: int):
        self.fs = mfd.fs
        self.mfd = mfd
        self.group = group
        self.dcn = mfd.gfd_pointer_map[group]

    @classmethod
    def read(cls, mfd: "MFD", group: int, cache: t.Optional[RTFSBlockCache] = None) -> "GFD":
        if cache is None:
            cache = mfd.fs.new_cache()
        self = GFD(mfd, group)
        buffer = cache.read_block(dcn=self.dcn + LABEL_BLOCK_OFFSET)
        self.read_mfd_cluster_map(buffer)
        buffer = cache.read_block(dcn=self.dcn + UFD_POINTER_BLOCK_OFFSET)
        self.ufd_pointer_map = list(struct.unpack_from(UFD_POINTER_BLOCK_FORMAT, buffer, 0))
        buffer = cache.read_block(dcn=self.dcn + GFD_NAME_ENTRY_BLOCK_OFFSET)
        self.name_entry_pointer_map = [Link(mfd.fs, x) for x in struct.unpack_from(UFD_POINTER_BLOCK_FORMAT, buffer, 0)]
        return self

    def read_mfd_cluster_map(self, buffer: bytes) -> None:
        """
            Read MFD cluster map.
            The MFD cluster map contains pointer to each cluster in the MFD.

            +-------------------------------------+
         0  |          MFD cluster size           |
            +-------------------------------------+
         2  |       DCN of MFD cluster 0          |
            +-------------------------------------+
            |                 ...                 |
            +-------------------------------------+
        16  |       DCN of MFD cluster 6          |
            +-------------------------------------+

            http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 22

        """
        blockette = struct.unpack_from(BLOCKETTE_FORMAT, buffer, CLUSTER_MAP_POS)
        self.gfd_cluster_size = blockette[0]
        self.gfd_cluster_map = list(blockette[1:])

    def read_gfd_name_entries(self, cache: t.Optional[RTFSBlockCache] = None) -> t.Iterator["MFDNameEntry"]:
        if cache is None:
            cache = self.fs.new_cache()
        for user, link in enumerate(self.name_entry_pointer_map):
            if not link.is_null:
                buffer = cache.read_block(link.to_lbn(self.gfd_cluster_map))
                yield MFDNameEntry.read(self.fs, buffer, MFD_ENTRY_LEN * link.entry)

    def read_dir_entries(
        self,
        ppn: t.Optional[PPN] = None,
        cache: t.Optional[RTFSBlockCache] = None,
    ) -> t.Iterator["UFDNameEntry"]:
        if ppn is not None and ppn.group != ANY_GROUP and ppn.group != self.group:
            return
        if cache is None:
            cache = self.fs.new_cache()
        for user, ufd_pointer in enumerate(self.ufd_pointer_map):
            if ufd_pointer != 0 and (ppn is None or ppn.user == ANY_USER or ppn.user == user):
                ufd_ppn = PPN(group=self.group, user=user)
                buffer = cache.read_block(dcn=ufd_pointer)
                ufd_label = UFDLabelEntry.read(self.fs, buffer)
                yield from self.fs.read_ufd_name_entries(ufd_label.ulnk, ufd_pointer, ppn=ufd_ppn, cache=cache)


class MFD:
    """
    Master File Directory (RDS1.1 or later)

    http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 25
    """

    fs: "RSTSFilesystem"
    gfd_pointer_map: t.List[int]  # Pointers to Group File Directories (RDS1.1)
    mdcn: int  # DCN of MFD

    def __init__(self, fs: "RSTSFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "RSTSFilesystem", mdcn: int) -> "MFD":
        self = MFD(fs)
        self.mdcn = mdcn

        # Read MFD label block
        buffer = self.fs.read_block(dcn=self.mdcn + LABEL_BLOCK_OFFSET)
        self.fs.read_mfd_cluster_map(buffer)  # Read cluster map from MFD label block

        # Read the GFD pointer block
        buffer = self.fs.read_block(dcn=self.mdcn + GFD_POINTER_BLOCK_OFFSET)
        self.gfd_pointer_map = list(struct.unpack_from(GFD_POINTER_BLOCK_FORMAT, buffer, 0))
        return self

    def read_gfds(self, ppn: t.Optional[PPN] = None, cache: t.Optional[RTFSBlockCache] = None) -> t.Iterator["GFD"]:
        """
        Read GFDs (Group File Directories)
        """
        if cache is None:
            cache = self.fs.new_cache()
        for group, dcn in enumerate(self.gfd_pointer_map):
            if dcn != 0 and (ppn is None or ppn.group == ANY_GROUP or ppn.group == group):
                yield GFD.read(self, group, cache)


class RSTSFilesystem(AbstractRXBlockFilesystem):
    """
    RSTS/E Filesystem

    RDS0 directory structure:

    MFD Label ---> MFD Name ---> MFD Name --> ...
                      |
                      |
                     \|/
                  UFD Label ---> UFD Name --> UFD Name --> ...
                                    |
                                   \|/
                               UFD Account
                                    |
                                   \|/
                                Retrieval --> Retrieval --> ...

    RDS1.x directory structure:

    MFD Label ---> MFD
                    |
                    +---> GFD ---> UFD Name --> UFD Name --> ...
                    |      |
                    |      +----> UFD Name --> UFD Name --> ...
                    |      |         |
                    |      |        \|/
                    |      |    UFD Account
                    |      |         |
                    |      |        \|/
                    |      |     Retrieval --> Retrieval --> ...
                    |      |
                    |      +----> UFD Name --> UFD Name --> ...
                    |      ...
                    |
                    +---> GFD ---> UFD Name --> UFD Name --> ...
                    ...

    """

    fs_name = "rsts"
    fs_description = "PDP-11 RSTS/E"
    fs_platforms = ["pdp11"]

    ppn: PPN  # Current Project Programmer Number

    device_size: int  # Device size in blocks
    dcs: int  # Device Cluster Size
    plvl: int  # Revision level (RDS1.1)
    ppcs: int  # Pack cluster size
    pstat: int  # Pack status
    pckid: str  # Pack ID
    structure_level: str  # RDS level
    mfd_cluster_size: int  # MFD cluster size
    mfd_cluster_map: t.List[int]  # MFD cluster map (DCN of MFD clusters 0 - 6)
    mfd_first_name_entry: Link  # Link to first name entry in MFD (RDS0)
    mfd: t.Optional[MFD] = None  # Master File Directory (RDS1.1 or later)

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "RSTSFilesystem":
        self = cls(file_or_dev)
        self.read_disk_pack_label()
        self.ppn = DEFAULT_PPN
        if strict:
            # Check if the SATT.SYS file exists
            try:
                self.get_file_entry(SAT_FILENAME)
            except FileNotFoundError:
                raise OSError(errno.EIO, "SATT.SYS file not found")
        return self

    def new_cache(self) -> "RTFSBlockCache":
        return RTFSBlockCache(self)

    def compute_dcs(self) -> int:
        """
        Compute DCS (Device Cluster Size)
        It is calculated such that all clusters on the disk
        can be specified by a 16-bit number
        """
        d = (self.device_size - 1) >> 16
        dcs = 1
        while d:
            d >>= 1
            dcs <<= 1
        return dcs

    def dcn_to_lbn(self, dcn: int) -> int:
        """
        Convert DCN (Device Cluster Number)
        to LBN (Logical Block Number)
        """
        return dcn * self.dcs

    def read_disk_pack_label(self) -> None:
        """Read disk pack label"""
        self.device_size = self.get_size() // BLOCK_SIZE
        self.dcs = self.compute_dcs()

        buffer = self.read_block(dcn=DISK_PACK_LABEL_DCN)
        (
            self.ulnk,  #   2 bytes  Link to first name entry in MFD (RDS0)
            _,  #           2 bytes  Unused (-1)
            mdcn,  #        2 bytes  DCN of MFD (RDS1.1)
            self.plvl,  #   2 bytes  Revision level (RDS1.1)
            self.ppcs,  #   2 bytes  Pack cluster size
            self.pstat,  #  2 bytes  Pack status
            pckid0,  #      2 bytes  Pack ID first word
            pckid1,  #      2 bytes  Pack ID second word
        ) = struct.unpack_from(DISK_PACK_LABEL_FORMAT, buffer, 0)
        self.pckid = rad50_word_to_asc(pckid0) + rad50_word_to_asc(pckid1)
        self.mfd_first_name_entry = Link(self, self.ulnk)

        if self.pstat & RDS1_FLAGS:  # RDS1.x
            if self.plvl == RDS11_PLVL:
                self.structure_level = "RDS1.1"
            else:
                self.structure_level = "RDS1.2"
            self.mfd = MFD.read(self, mdcn)
        else:  # RDS0
            self.plvl = RDS0_PLVL  # Set RDS level to 0
            self.structure_level = "RDS0"
            self.read_mfd_cluster_map(buffer)

    def read_mfd_cluster_map(self, buffer: bytes) -> None:
        """
            Read MFD cluster map.
            The MFD cluster map contains pointer to each cluster in the MFD.

            +-------------------------------------+
         0  |          MFD cluster size           |
            +-------------------------------------+
         2  |       DCN of MFD cluster 0          |
            +-------------------------------------+
            |                 ...                 |
            +-------------------------------------+
        16  |       DCN of MFD cluster 6          |
            +-------------------------------------+

            http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 22

        """
        blockette = struct.unpack_from(BLOCKETTE_FORMAT, buffer, CLUSTER_MAP_POS)
        self.mfd_cluster_size = blockette[0]
        self.mfd_cluster_map = list(blockette[1:])

    def read_ufd_cluster_map(self, buffer: bytes) -> t.List[int]:
        """
            Read UFD cluster map.
            The UFD cluster map contains pointer to each cluster in the UFD.

            +-------------------------------------+
         0  |          UFD cluster size           |
            +-------------------------------------+
         2  |       DCN of UFD cluster 0          |
            +-------------------------------------+
            |                 ...                 |
            +-------------------------------------+
        16  |       DCN of UFD cluster 6          |
            +-------------------------------------+

            http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 38

        """
        blockette = struct.unpack_from(BLOCKETTE_FORMAT, buffer, CLUSTER_MAP_POS)
        return list(blockette[1:])

    def read_mfd_name_entries(self, cache: t.Optional[RTFSBlockCache] = None) -> t.Iterator["MFDNameEntry"]:
        """
        Read MFD name entries
        http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 20
        """
        if cache is None:
            cache = self.new_cache()
        if self.mfd is not None:  # RDS1.x
            for gfd in self.mfd.read_gfds(cache=cache):
                yield from gfd.read_gfd_name_entries(cache=cache)
        else:  # RDS0
            link = self.mfd_first_name_entry
            while not link.is_null:
                buffer = cache.read_block(link.to_lbn(self.mfd_cluster_map))
                mfd_entry = MFDNameEntry.read(self, buffer, MFD_ENTRY_LEN * link.entry)
                link = mfd_entry.ulnk
                if mfd_entry.is_mfd_name_entry:
                    yield mfd_entry

    def read_ufd_name_entries(
        self,
        link: Link,
        ufd_uar: int,
        ppn: PPN,
        cache: t.Optional[RTFSBlockCache] = None,
    ) -> t.Iterator["UFDNameEntry"]:
        """
        Read UFD name entries
        http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 20
        """
        if cache is None:
            cache = self.new_cache()
        buffer = cache.read_block(link.block + ufd_uar)
        ufd_cluster_map = self.read_ufd_cluster_map(buffer)
        while not link.is_null:
            buffer = cache.read_block(link.to_lbn(ufd_cluster_map))
            ufd_entry = UFDNameEntry.read(self, ppn, ufd_uar, buffer, UFD_ENTRY_LEN * link.entry)
            buffer = cache.read_block(ufd_entry.uaa.to_lbn(ufd_cluster_map))
            ufd_entry.account_entry = UFDAccountEntry.read(self, buffer, UFD_ENTRY_LEN * ufd_entry.uaa.entry)
            link = ufd_entry.ulnk
            yield ufd_entry

    def read_ufd_label_entry(self, dcn: int) -> "UFDLabelEntry":
        """
        Read UFD label entry
        http://elvira.stacken.kth.se/rstsdoc/rsts-doc-v80/extra/mayfieldRSTS8internals.pdf Pag 33
        """
        buffer = self.read_block(dcn=dcn)
        return UFDLabelEntry.read(self, buffer)

    def read_block(
        self,
        block_number: int = 0,
        number_of_blocks: int = 1,
        dcn: t.Optional[int] = None,
    ) -> bytes:
        if dcn is not None:
            block_number = self.dcn_to_lbn(dcn)
        return super().read_block(block_number, number_of_blocks)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def read_dir_entries(self, ppn: PPN, cache: t.Optional[RTFSBlockCache] = None) -> t.Iterator["UFDNameEntry"]:
        if cache is None:
            cache = self.new_cache()
        if self.mfd is not None:  # RDS1.x
            for gfd in self.mfd.read_gfds(ppn=ppn, cache=cache):
                yield from gfd.read_dir_entries(ppn=ppn, cache=cache)
        else:  # RDS0
            for mfd_entry in self.read_mfd_name_entries(cache):
                match = True
                if ppn.group != ANY_GROUP:
                    match &= mfd_entry.ppn.group == ppn.group
                if ppn.user != ANY_USER:
                    match &= mfd_entry.ppn.user == ppn.user
                if match:
                    if mfd_entry.ppn == ACCOUNT_1_1_PPN:
                        for ufd_entry in self.read_ufd_name_entries(
                            mfd_entry.ulnk, mfd_entry.uar, ppn=mfd_entry.ppn, cache=cache
                        ):
                            if ufd_entry.is_ufd_name_entry:
                                yield ufd_entry
                    elif mfd_entry.uar != 0:
                        ufd_label = self.read_ufd_label_entry(mfd_entry.uar)
                        yield from self.read_ufd_name_entries(
                            ufd_label.ulnk,
                            mfd_entry.uar,
                            ppn=mfd_entry.ppn,
                            cache=cache,
                        )

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
        ppn: t.Optional[PPN] = None,
    ) -> t.Iterator["UFDNameEntry"]:
        if ppn is None:
            ppn = self.ppn
        ppn, pattern = rsts_split_fullname(fullname=pattern, wildcard=wildcard, ppn=ppn)
        for entry in self.read_dir_entries(ppn=ppn):
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    @property
    def entries_list(self) -> t.Iterator[UFDNameEntry]:
        yield from self.read_dir_entries(ppn=self.ppn)

    def get_file_entry(self, fullname: str) -> UFDNameEntry:
        fullname = rsts_canonical_fullname(fullname)
        if not fullname:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        ppn, basename = rsts_split_fullname(fullname=fullname, wildcard=False, ppn=self.ppn)
        try:
            return next(self.filter_entries_list(basename, wildcard=False, ppn=ppn))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[UFDNameEntry]:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        if options.get("uic"):
            # Listing of all PPN
            for mfd_entry in self.read_mfd_name_entries():
                sys.stdout.write(f"{mfd_entry.ppn}\n")
            return
        files = 0
        blocks = 0
        ppn, pattern = rsts_split_fullname(fullname=pattern, wildcard=True, ppn=self.ppn)
        if not options.get("brief"):
            sys.stdout.write(f" Name .Ext  Size    Prot   Date       {volume_id}:{ppn}\n")
        for x in self.filter_entries_list(pattern, ppn=ppn, include_all=True, wildcard=True):
            if options.get("brief"):
                # Lists only file names and file types
                sys.stdout.write(f"{x.filename:<6}.{x.extension:<3}\n")
            else:
                date = x.creation_date.strftime("%d-%b-%y %H:%M")
                sys.stdout.write(
                    f"{x.filename:<6}.{x.extension:<3} {x.account_entry.usiz:>5}   <{x.uprot:>3}> {date}\n"
                )
                blocks += x.account_entry.usiz
                files += 1
        if options.get("brief"):
            return
        sys.stdout.write("\n")
        sys.stdout.write(f" Total of {blocks} blocks in {files} files in {volume_id}:{ppn}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        ppn = None
        if arg and "[" in arg:
            try:
                ppn = PPN.from_str(arg)
                arg = arg.split("]", 1)[1]
            except Exception:
                return
        if ppn is not None:
            for mfd_entry in self.read_mfd_name_entries():
                match = True
                if ppn.group != ANY_GROUP:
                    match &= mfd_entry.ppn.group == ppn.group
                if ppn.user != ANY_USER:
                    match &= mfd_entry.ppn.user == ppn.user
                if match:
                    sys.stdout.write(f"{ppn}\n")
                    sys.stdout.write(dump_struct(mfd_entry.__dict__))
                    sys.stdout.write("\n\nUDF Entries Label\n\n")
                    if mfd_entry.ppn == ACCOUNT_1_1_PPN and self.structure_level == "RDS0":
                        for ufd_entry in self.read_ufd_name_entries(mfd_entry.ulnk, mfd_entry.uar, mfd_entry.ppn):
                            if ufd_entry.is_ufd_name_entry:
                                sys.stdout.write(f"{ufd_entry}\n")
                    else:
                        ufd_label = self.read_ufd_label_entry(mfd_entry.uar)
                        for ufd_entry in self.read_ufd_name_entries(ufd_label.ulnk, mfd_entry.uar, mfd_entry.ppn):
                            sys.stdout.write(f"{ufd_entry}\n")

        else:
            sys.stdout.write("Disk Pack Label\n\n")
            sys.stdout.write(dump_struct(self.__dict__))
            sys.stdout.write("\n\nMFD Cluster Map\n\n")
            for i, dcn in enumerate(self.mfd_cluster_map):
                if dcn:
                    sys.stdout.write(f"Cluster {i} -> DCN {dcn}\n")
            if self.mfd is not None:
                sys.stdout.write("\nGFD Pointer Map\n\n")
                for i, dcn in enumerate(self.mfd.gfd_pointer_map):
                    if dcn:
                        sys.stdout.write(f"Cluster {i} -> DCN {dcn}\n")
            # Listing of all PPN
            sys.stdout.write("\nProject Programmer Numbers\n\n")
            for mfd_entry in self.read_mfd_name_entries():
                sys.stdout.write(f"{mfd_entry}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def chdir(self, fullname: str) -> bool:
        """
        Change the current Project Programmer Number
        """
        try:
            self.ppn = PPN.from_str(fullname)
            return True
        except Exception:
            return False

    def get_pwd(self) -> str:
        return str(self.ppn)



from xferx.shell import Shell
from xferx.unix.unix6fs import UNIXFilesystem

DSK = "tests/dsk/unixv6.dsk"


def test_unix6_read():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /unix6 {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, UNIXFilesystem)
    assert fs.version == 6

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir t:/", batch=True)
    shell.onecmd("dir t:/etc/", batch=True)
    shell.onecmd("type t:/etc/passwd", batch=True)

    x = fs.read_text("1")
    assert x.startswith("1\n")

    l = list(fs.entries_list)
    filenames = [x.filename for x in l if not x.is_empty]
    assert "1140k" in filenames

    entry = fs.get_file_entry("/1140k")
    assert entry.inode.is_large
    assert entry.inode.is_huge
    assert entry.inode.addr[-1] != 0

    entry = fs.get_file_entry("/95k")
    assert entry.inode.is_large
    assert not entry.inode.is_huge
    assert entry.inode.addr[-1] == 0

    entry = fs.get_file_entry("/9k")
    assert entry.inode.is_large
    assert not entry.inode.is_huge
    assert entry.inode.addr[-1] == 0

    entry = fs.get_file_entry("/etc/passwd")
    assert not entry.inode.is_large
    assert not entry.inode.is_huge
    assert entry.inode.addr[-1] == 0



import errno
import typing as t

from ..abstract import AbstractBlockFilesystem, AbstractFile
from ..device.abstract import AbstractDevice
from ..device.block import BlockDevice
from ..device.rx import RXBlockDevice

__all__ = [
    "AbstractRXBlockFilesystem",
]


class AbstractRXBlockFilesystem(AbstractBlockFilesystem):
    """Abstract base class for block-based filesystems with RX01/RX02 support."""

    dev: BlockDevice  # BlockDevice or RXBlockDevice

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = RXBlockDevice(file_or_device)
        elif isinstance(file_or_device, BlockDevice):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, "Not a valid block device")
